# Copyright © The University of Edinburgh, 2023.
# Development has been supported by GSK.

from phenonaut.data import Dataset
from phenonaut.transforms.transformer import Transformer
import numpy as np

# Inserted as an example of making a custom transformer from a simple function
log2_transformer = Transformer(np.log2)

# Alternatively, the class
class Log2(Transformer):
    def __init__(self):
        super().__init__(np.log2, new_feature_names="log2_")



# Copyright © The University of Edinburgh, 2023.
# Development has been supported by GSK.

import pandas as pd
from phenonaut.data import Dataset
import numpy as np
from phenonaut.phenonaut import Phenonaut
from typing import Optional, Union, Callable
import numpy as np
from scipy.stats import spearmanr
from random import sample
from sklearn.metrics import pairwise_distances
from joblib import Parallel, delayed
from warnings import warn


def percent_replicating(
    ds: Union[Dataset, Phenonaut, pd.DataFrame],
    features: Optional[list[str]] = None,
    perturbation_identifier: Optional[str] = None,
    groupby: Optional[list[str]] = None,
    n_iters: int = 1000,
    phenotypic_metric: Union[str, Callable] = "spearman",
    min_number_of_repeats: int = 2,
    return_full_performance_df: bool = False,
    n_jobs: int = -1,
):
    """Calculate percent replicating

    Percent replicating is defined by Way et. al. in:
    Way, Gregory P., et al. "Morphology and gene expression profiling provide complementary
    information for mapping cell state." Cell systems 13.11 (2022): 911-923.
    or on bioRxiv:
    https://www.biorxiv.org/content/10.1101/2021.10.21.465335v2

    For each query compound at a concentration and in a common well (multiple plates), calculate
    pairwise distances and obtain the median.

    Next we compare this against the 95th percentile of a null distribution.
    If the replicates are greater, then it is replicating.

    To create the null distribution, we remove the query compound from the pool of possible
    compounds (at any concentration and in any well). Then we pick the same number of compounds
    as there were repeats of the query compound at a given concentration and in a well. We
    perform this to match the cardinality of repeats across distributions. From the data,
    one profile for each non-matching compound is chosen and then the median recorded when
    compared with a phenoytypic metric (by default Spearman's rank). This creation and scoring
    of the null distribution is performed (by default) 1,000 times.

    The original mean score of the query compound repeats is then compared against the 95th
    percentile of this null distribution. If it is greater, then compound is deemed replicating.

    The percent replicating is then calculated from the number which were replicating versus the
    number which were not.

    As the calculation is demanding, the function makes use of the joblib library for parallel
    calculation of the null distribution.

    Parameters
    ----------
    ds : Union[Dataset, Phenonaut, pd.DataFrame],
        Input data in the form of a Phenonaut dataset, a Phenonaut object, or a pd.DataFrame
        containing profiles of perturbations. If a Phenonaut object is supplied, then the last added
        Dataset will be used.
    features : features:Optional[list[str]]=None,
        Features list which capture the phenotypic responses to perturbations. Only required if
        a pd.DataFrame is supplied in place of the ds argument.  If a Phenonaut object, or Dataset,
        the features will be examined and used as present in the Dataset/Phenonaut object last
        added Dataset. By default None.
    perturbation_identifier : Optional[str]
        Required if a pd.DataFrame is supplied in place of the ds argument. If a Phenonaut object is
        supplied, then the perturbation_identifier is obtained from the perturbation_column property
        of the Dataset. The same happens if a Dataset is supplied. By default, None.
    groupby : Optional[list[str]], optional
        Columns in the dataframe that should be used to group perturbations in addition to the
        perturbation_identifier argument. To replicate percent replicating as defined by Way,
        the column names which capture concentration and well position should be supplied here.
        Way does however define variations of the percent replicating metric, where concentration
        and well position are relaxed. If performing the calculation in this manner, then groupby
        may be None.
    n_iters : int, optional
        Number of times the non-matching compound replicates should be sampled to compose the null
        distribution, by default 1000
    phenotypic_metric : Union[str, Callable], optional
        A callable phenotypic metric may be passed in, or the special case of the string 'spearman'
        which returns the rank score. Other valid and usable text strings may be used which are then
        passed to scikit's pairwise_distances function.  See documentation for usable strings
        defining metrics. By default 'spearman'
    min_number_of_repeats : int, optional
        The minimum number of matched query compounds which must be present to include the treatment
        in the percent replicating summary statistic. By default 2.
    return_full_performance_df : bool
        If True, then a tuple is returned with the percent replicating score in the first position,
        and a pd.DataFrame containing full information on each repeat
    n_jobs : int, optional
        The n_jobs argument is passed to joblib for parallel execution and defines the number of
        cpus to use.  A value of -1 denotes that the system should determine how many jobs to run.
        By default -1.

    Returns
    -------
    Union[float, tuple[float, pd.DataFrame]]
        If return_full_performance_df is False, then only the percent replicating score is returned.
        If True, then a tuple is returned, with percent replicating in the first position, and a
        pd.DataFrame in the second position containing the median repeat scores, as well as median
        null distribution scores in an easy to analyse format.

    """
    if isinstance(ds, Phenonaut):
        ds = ds[-1]

    if isinstance(ds, Dataset):
        df = ds.df
        features = ds.features
        if perturbation_identifier is None:
            if ds.perturbation_column is None:
                raise ValueError(
                    "perturbation_identifier was not supplied, and perturbation_column is not set in the dataset"
                )
            perturbation_identifier = ds.perturbation_column
    elif isinstance(ds, pd.DataFrame):
        df = ds
        if features is None:
            raise ValueError(
                "Must supply features if using a pd.DataFrame to calculate % replicating"
            )
        if perturbation_identifier is None:
            raise ValueError(
                "Must supply perturbation_identifier if using a pd.DataFrame to calculate % replicating"
            )

    if groupby is None:
        groupby = []
    if return_full_performance_df:
        perf_dict = {
            perturbation_identifier: [],
            **{g: [] for g in groupby},
            "median_replicate_score": [],
            "null_95th_percentile": [],
            "is_replicating": [],
            **{f"null_{i+1}": [] for i in range(n_iters)},
        }

    # Private function which will be executed in parallel by joblib
    # Returns a tuple of
    # - median_replicate_score
    # - median_non_replicate_scores
    def __get_cpd_pct_replicating_info(
        df: pd.DataFrame,
        features: list[str],
        g_df: pd.DataFrame,
        g_name: Union[tuple[Union[str, float]], str],
    ):
        # If only grouped by one thing, like a string, then the group 'name' variable is just a string, not a tuple
        # containing the multi index keys.  Therefore we need to properly extract the name
        query_cpd_name = g_name[0] if isinstance(g_name, tuple) else g_name
        cardinality = g_df.shape[0]
        if cardinality < min_number_of_repeats:
            return None

        median_replicate_score = np.median(
            pairwise_distances(g_df[features], metric=phenotypic_metric)[
                np.tril_indices(cardinality, k=-1)
            ]
        )
        median_non_replicate_scores = np.empty(n_iters)
        non_matching_compound_choices = [p for p in perturbation_ids if p != query_cpd_name]

        # If testing a very small pool of compounds with:
        # DMSO x 10
        # Cpd1 x 2
        # Cpd2 x 2
        # Then we can never match the cardinality of the DMSO pool's 10x repeats. Warn about this and then reduce cardinality to 2.
        if cardinality > len(non_matching_compound_choices):
            warn(
                f"Less non-matching compounds than cardinality (repeats of matching treatment), reducing cardinality from {cardinality} to {len(non_matching_compound_choices)}"
            )
            cardinality = len(non_matching_compound_choices)

        for i in range(n_iters):
            sampled_choices = sample(non_matching_compound_choices, k=cardinality)
            null_data = np.vstack(
                [
                    df.query(f"{perturbation_identifier}==@non_matching_perturbation")[features]
                    .sample(1)
                    .values
                    for non_matching_perturbation in sampled_choices
                ]
            )
            median_non_replicate_scores[i] = np.median(
                pairwise_distances(null_data, metric=phenotypic_metric)[
                    np.tril_indices(cardinality, k=-1)
                ]
            )
        return (
            g_name,
            g_df.shape[0],
            cardinality,
            median_replicate_score,
            median_non_replicate_scores,
        )

    samples_replicated = []
    perturbation_ids = df[perturbation_identifier].unique().tolist()

    # spearmanr returns (rank, p-val) tuple, we only want the rank if spearman specified.
    if phenotypic_metric == "spearman":
        phenotypic_metric = lambda x, y: spearmanr(x, y)[0]

    cpd_pct_replicating_info = Parallel(n_jobs=n_jobs)(
        delayed(__get_cpd_pct_replicating_info)(df, features, g_df, g_name)
        for g_name, g_df in df.groupby(
            perturbation_identifier if not groupby else [perturbation_identifier] + groupby
        )
    )

    for t in cpd_pct_replicating_info:
        if t is None:
            continue
        percentile_score = np.percentile(t[4], 95)
        if t[3] > percentile_score:
            samples_replicated.append(1)
        else:
            samples_replicated.append(0)
        if return_full_performance_df:
            perf_dict[perturbation_identifier].append(t[0][0] if isinstance(t[0], tuple) else t[0])
            for gb_i, gb in enumerate(groupby):
                perf_dict[gb].append(t[0][gb_i])
            perf_dict["median_replicate_score"].append(t[3])
            perf_dict["null_95th_percentile"].append(percentile_score)
            perf_dict["is_replicating"].append(samples_replicated[-1])
            for i in range(n_iters):
                perf_dict[f"null_{i+1}"].append(t[4][i])
    percent_replicating_score = (np.sum(samples_replicated) / len(samples_replicated)) * 100
    if return_full_performance_df:
        return (percent_replicating_score, pd.DataFrame(perf_dict))
    else:
        return percent_replicating_score



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from collections import namedtuple
from pathlib import Path
import pathlib
from typing import List, NamedTuple, Optional, Union
from ..data import Dataset
import pandas as pd
from .base import PackagedDataset
import gzip, tarfile
import shutil
import pandas as pd
import numpy as np
import h5py
import datetime


class LINCS_Cell_Painting(PackagedDataset):
    """LINCS Cell Painting Dataset - https://clue.io/

    This PackagedDataset provides supplies the following pd.DataFrames
    (queryable by calling the inherited ".keys" method):

    Parameters
    ----------
    root : Union[Path, str]
        Local directory containing the prepared dataset. If the dataset is
        not found here and the argument download=True is not given, then an
        error is raised. If download=True and the processed dataset is
        absent, then it is downloaded the directory pointed at by the
        'raw_data' argument detailed below. If raw_data_dir is a
        non-absolute path, such as a single directory, then it is created as
        a subdirectory of this root directory.
    download : bool, optional
        If true and the processed dataset is not found in the root
        directory, then the dataset is downloaded and processed.
        By default False.
    raw_data_dir : Optional[Union[Path, str]], optional
        If downloading and preparing the dataset, then a directory for the
        raw data may be specified. If a non-absolute location is given,
        then it is created in a subdirectory of the root directory specified
        as the first argument. Absolute paths may be used to place raw data
        files and intermediates in another location, such as scratch disks
        etc, by default Path("raw_data").
    """

    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
    ):
        """LINCS Cell Painting Dataset - https://clue.io/

        This PackagedDataset provides supplies the following pd.DataFrames
        (queryable by calling the inherited ".keys" method):

        Parameters
        ----------
        root : Union[Path, str]
            Local directory containing the prepared dataset. If the dataset is
            not found here and the argument download=True is not given, then an
            error is raised. If download=True and the processed dataset is
            absent, then it is downloaded the directory pointed at by the
            'raw_data' argument detailed below. If raw_data_dir is a
            non-absolute path, such as a single directory, then it is created as
            a subdirectory of this root directory.
        download : bool, optional
            If true and the processed dataset is not found in the root
            directory, then the dataset is downloaded and processed.
            By default False.
        raw_data_dir : Optional[Union[Path, str]], optional
            If downloading and preparing the dataset, then a directory for the
            raw data may be specified. If a non-absolute location is given,
            then it is created in a subdirectory of the root directory specified
            as the first argument. Absolute paths may be used to place raw data
            files and intermediates in another location, such as scratch disks
            etc, by default Path("raw_data").
        """

        super().__init__(root, raw_data_dir)
        self.name = "LINCS Cell Painting "
        self.processed_h5_file = self.root / "LINCS_CellPainting_phenonaut.h5"
        # If the dataset is missing, get it
        self._call_if_file_missing(self.processed_h5_file, self._make, None)

        store = pd.HDFStore(self.processed_h5_file)

        self.register_ds_key("ds")

        store.close()

    def _make(self, remove_intermediates=False):
        """Make main LINCS CellPainting HDF5 file

        Downloads the two compressed data files required (Batches 1 and 2),
        decompresses it, extracts the data into a pd.DataFrame, merges the
        sig_info data file which includes data on the perturbation type.

        Parameters
        ----------
        remove_intermediates : bool, optional
            If true, then the downloaded archive is removed, by default False
        """
        # XXX TODO change remove_intermediates to default to True.
        # XXX TODO change compression to 9.

        # Make sure the required files to make the dataset exist
        DownloadableFileInfo = namedtuple(
            "DownloadableFileInfo", ["filename", "compressed_filename", "url"]
        )
        required_files = [
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt",
                "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
            ),
        ]
        for req_file in required_files:
            self._call_if_file_missing(
                self.root / req_file.filename,
                self._download,
                {
                    "source": req_file.url,
                    "destination": self.root / req_file.compressed_filename,
                    "extract": True,
                },
            )

        # Make sure the large GCTX file is present
        CMAP_GCTX_GZ_URL = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_Level5_COMPZ_n118050x12328_2017-03-06.gctx.gz"
        CMAP_GCTX_GZ_FILENAME = "GSE70138_Broad_LINCS_Level5_COMPZ_n118050x12328_2017-03-06.gctx"
        self._call_if_file_missing(
            self.root / CMAP_GCTX_GZ_FILENAME,
            self._download,
            {
                "source": CMAP_GCTX_GZ_URL,
                "destination": self.root / (CMAP_GCTX_GZ_FILENAME + ".gz"),
                "extract": True,
            },
        )

        # Load in CMAP large dataset from GCTX file, reading it as an HDF5 file, and then walking the correct path.
        f = h5py.File(self.root / CMAP_GCTX_GZ_FILENAME)
        print(f"{datetime.datetime.now()} Extracting large matrix... ")
        self._df = pd.DataFrame(
            data=f["0"]["DATA"]["0"]["matrix"],
            columns=f["0"]["META"]["ROW"]["id"],
            index=f["0"]["META"]["COL"]["id"],
        )
        self._df.index = self._df.index.map(lambda x: x.decode("utf8"))
        self._df.columns = self._df.columns.map(lambda x: x.decode("utf8"))
        print(f"{datetime.datetime.now()} ..done")

        # Here, we generate cmap_phenonaut.h5, containing a merged df
        print(f"{datetime.datetime.now()} Merging")
        self._sig_info = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )
        pert_dose = (
            self._sig_info["pert_idose"].apply(lambda x: x.replace(" um", "").strip()).astype(float)
        )
        self._sig_info["pert_idose"] = np.where(pert_dose < 0, np.nan, pert_dose)
        self._df = self._df.merge(self._sig_info, left_index=True, right_index=True, copy=False)

        h5_store = pd.HDFStore(self.processed_h5_file, "w", complevel=0)
        h5_store["df"] = self._df

        h5_store["gene_info"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )
        h5_store["sig_metrics"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )
        h5_store["pert_info"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )

        h5_store["inst_info"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )

        h5_store["creation_date"] = pd.Series(str(datetime.datetime.now()))
        h5_store.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        if key == "landmark_gene_ids":
            store = pd.HDFStore(self.processed_h5_file)
            df = store["/gene_info"]
            return df.query("pr_is_lm == 1").index.to_numpy(dtype=str)

        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        if key == "ds":
            key = "df"
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        features = [
            f
            for f in list(df.columns)
            if f
            not in [
                "pert_id",
                "pert_iname",
                "pert_type",
                "cell_id",
                "pert_idose",
                "pert_itime",
                "distil_id",
            ]
        ]
        ds = Dataset("cmap", df, {"features": features})
        store.close()
        return ds



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from mvlearn.semi_supervised import CTClassifier
from phenonaut.predict.predictor_dataclasses import (
    PhenonautPredictor,
    HyperparameterInt,
    HyperparameterFloat,
    HyperparameterCategorical,
    HyperparameterLog,
)

default_classifiers = [
    PhenonautPredictor("Naive Bayes", GaussianNB),
    PhenonautPredictor(
        "Random Forest",
        RandomForestClassifier,
        [
            HyperparameterCategorical("max_depth", (5, 10, 20, None)),
            HyperparameterInt("n_estimators", 50, 200),
            HyperparameterCategorical("max_features", (None, "auto", "log2")),
        ],
        constructor_kwargs={"n_jobs": -1},
    ),
    PhenonautPredictor(
        "KNN",
        KNeighborsClassifier,
        HyperparameterInt("n_neighbors", 2, 10),
        max_optuna_trials=9,
        constructor_kwargs={"n_jobs": -1},
    ),
    PhenonautPredictor(
        "Linear SVM",
        SVC,
        [
            HyperparameterCategorical("kernel", ("linear",)),
            HyperparameterLog("C", 0.001, 10),
            HyperparameterCategorical("gamma", ("scale", "auto")),
        ],
        dataset_size_cutoff=1000,
    ),
    PhenonautPredictor(
        "RBF SVM",
        SVC,
        [
            HyperparameterCategorical("kernel", ("rbf",)),
            HyperparameterLog("C", 0.001, 10),
            HyperparameterCategorical("gamma", ("scale", "auto")),
        ],
        dataset_size_cutoff=1000,
    ),
    PhenonautPredictor(
        "Gaussian Process",
        GaussianProcessClassifier,
        constructor_kwargs={
            "copy_X_train": False,
            "n_jobs": -1,
        },
        embed_in_results=False,
    ),
    PhenonautPredictor(
        "Decision Tree",
        DecisionTreeClassifier,
        [
            HyperparameterCategorical("criterion", ("gini", "entropy")),
            HyperparameterCategorical("max_depth", (None, 5, 10, 15, 20)),
            HyperparameterCategorical("max_features", (None, "auto", "log2")),
        ],
    ),
    PhenonautPredictor(
        "SciKit NeuralNet",
        MLPClassifier,
        [
            HyperparameterLog("alpha", 0.00001, 0.001),
            HyperparameterCategorical("solver", ["adam", "sgd"]),
            HyperparameterCategorical("max_iter", [1000]),
            HyperparameterCategorical("learning_rate", ["constant", "adaptive"]),
        ],
    ),
    PhenonautPredictor("AdaBoost", AdaBoostClassifier),
    PhenonautPredictor("QDA", QuadraticDiscriminantAnalysis),
    PhenonautPredictor("mvlearn CTClassifier (GNB,GNB)", CTClassifier, num_views=2, max_classes=2),
]



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from curses import meta
import pytest
import phenonaut
from phenonaut.data.dataset import Dataset
import pandas as pd
import numpy as np
import tempfile
from io import StringIO


from phenonaut.phenonaut import Phenonaut


def test_possible_dataset_combinations(phenonaut_object_iris_4_views):
    all_possible_views = phenonaut_object_iris_4_views.get_dataset_combinations()
    assert all_possible_views == (
        ("Iris_view1",),
        ("Iris_view2",),
        ("Iris_view3",),
        ("Iris_view4",),
        ("Iris_view1", "Iris_view2"),
        ("Iris_view1", "Iris_view3"),
        ("Iris_view1", "Iris_view4"),
        ("Iris_view2", "Iris_view3"),
        ("Iris_view2", "Iris_view4"),
        ("Iris_view3", "Iris_view4"),
        ("Iris_view1", "Iris_view2", "Iris_view3"),
        ("Iris_view1", "Iris_view2", "Iris_view4"),
        ("Iris_view1", "Iris_view3", "Iris_view4"),
        ("Iris_view2", "Iris_view3", "Iris_view4"),
        ("Iris_view1", "Iris_view2", "Iris_view3", "Iris_view4"),
    )


def test_possible_dataset_combinations_indexes(phenonaut_object_iris_4_views):
    all_possible_views = phenonaut_object_iris_4_views.get_dataset_combinations(return_indexes=True)
    assert all_possible_views == (
        (0,),
        (1,),
        (2,),
        (3,),
        (0, 1),
        (0, 2),
        (0, 3),
        (1, 2),
        (1, 3),
        (2, 3),
        (0, 1, 2),
        (0, 1, 3),
        (0, 2, 3),
        (1, 2, 3),
        (0, 1, 2, 3),
    )


def test_get_dataset_index_from_name(phenonaut_object_iris_4_views):
    assert phenonaut_object_iris_4_views.get_dataset_index_from_name(
        ["Iris_view4", "Iris_view3", "Iris_view2", "Iris_view1"]
    ) == [3, 2, 1, 0]


def test_phenonaut_obj_from_df():
    df = pd.DataFrame({"id": [9, 8, 7], "A": [1, 2, 3], "B": [4, 5, 6]})
    another_df = pd.DataFrame({"id2": [9, 8, 7], "A2": [1, 2, 3], "B": [4, 5, 6]})
    ph0 = phenonaut.Phenonaut(df, metadata={"features": ["B"]}, dataframe_name="Bob")
    assert ph0.ds.features == ["B"]
    assert ph0.ds.name == "Bob"
    ph0 = phenonaut.Phenonaut(df, metadata={"features_prefix": ["B"]})
    assert ph0.ds.features == ["B"]
    ph0 = phenonaut.Phenonaut(df, metadata={"features_prefix": ""})
    assert set(ph0.ds.features) == set(["B", "A", "id"])
    ph0 = phenonaut.Phenonaut([df, another_df], metadata={"features_prefix": ""})
    assert set(ph0.datasets[0].features) == set(["B", "A", "id"])
    assert set(ph0.datasets[1].features) == set(["B", "A2", "id2"])
    ph0 = phenonaut.Phenonaut([df, another_df], metadata={"features": ["B"]})
    assert set(ph0.datasets[0].features) == set(["B"])
    assert set(ph0.datasets[1].features) == set(["B"])
    ph0 = phenonaut.Phenonaut(
        [df, another_df],
        metadata=[{"features": ["B"]}, {"features_prefix": ""}],
        dataframe_name=["Bob", "Robert"],
    )
    assert set(ph0.datasets[0].features) == set(["B"])
    assert set(ph0.datasets[1].features) == set(["A2", "B", "id2"])
    assert ph0.datasets[0].name == "Bob"
    assert ph0.datasets[-1].name == "Robert"


def test_aggregate_dataset_in_ph0_obj():
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2

    As there are multiple fields of view, we merge on common ROW, COLUMN and BARCODE fields.
    Merging uses np.mean, but may be any callable which returns a single value

    """

    df = pd.DataFrame(
        {
            "ROW": [1, 1, 1, 1, 1, 1],
            "COLUMN": [1, 1, 1, 1, 2, 2],
            "BARCODE": ["Plate1", "Plate1", "Plate2", "Plate2", "Plate1", "Plate1"],
            "feat_1": [1.2, 1.3, 5.2, 6.2, 0.1, 0.2],
            "feat_2": [1.2, 1.4, 5.1, 6.1, 0.2, 0.2],
            "feat_3": [1.3, 1.5, 5, 6.8, 0.3, 0.38],
            "filename": [
                "fileA.png",
                "FileB.png",
                "FileC.png",
                "FileD.png",
                "fileE.png",
                "FileF.png",
            ],
            "FOV": [1, 2, 1, 2, 1, 2],
        }
    )

    phe = phenonaut.Phenonaut(df, metadata={"features_prefix": "feat_"})
    phe.aggregate_dataset(["ROW", "COLUMN", "BARCODE"])
    assert len(phe.datasets) == 2
    assert set(phe.ds.features) == set(["feat_1", "feat_2", "feat_3"])
    assert abs(phe.ds.df["feat_1"][0] - 1.25 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][1] - 5.70 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][2] - 0.15 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][0] - 1.3 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][1] - 5.6 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][2] - 0.2 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][0] - 1.4 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][1] - 5.9 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][2] - 0.34 < 0.00001)


def test_aggregate_dataset():
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2

    As there are multiple fields of view, we merge on common ROW, COLUMN and BARCODE fields.
    Merging uses np.mean, but may be any callable which returns a single value

    """
    df = pd.DataFrame(
        {
            "ROW": [1, 1, 1, 1, 1, 1],
            "COLUMN": [1, 1, 1, 1, 2, 2],
            "BARCODE": ["Plate1", "Plate1", "Plate2", "Plate2", "Plate1", "Plate1"],
            "feat_1": [1.2, 1.3, 5.2, 6.2, 0.1, 0.2],
            "feat_2": [1.2, 1.4, 5.1, 6.1, 0.2, 0.2],
            "feat_3": [1.3, 1.5, 5, 6.8, 0.3, 0.38],
            "filename": [
                "fileA.png",
                "FileB.png",
                "FileC.png",
                "FileD.png",
                "fileE.png",
                "FileF.png",
            ],
            "FOV": [1, 2, 1, 2, 1, 2],
        }
    )

    phe = phenonaut.Phenonaut([df, df])
    phe.aggregate_dataset(
        ["ROW", "COLUMN", "BARCODE"], datasets=[0, 1], new_names_or_prefix=["M1", "M2"]
    )
    assert len(phe.datasets) == 4
    phe = phenonaut.Phenonaut([df, df])
    phe.aggregate_dataset(
        ["ROW", "COLUMN", "BARCODE"],
        datasets=[0, 1],
        new_names_or_prefix=["M1", "M2"],
        inplace=True,
    )
    assert len(phe.datasets) == 2
    assert set(phe.ds.features) == set(["feat_1", "feat_2", "feat_3"])
    assert abs(phe.ds.df["feat_1"][0] - 1.25 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][1] - 5.70 < 0.00001)
    assert np.abs(phe.ds.df["feat_1"][2] - 0.15 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][0] - 1.3 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][1] - 5.6 < 0.00001)
    assert np.abs(phe.ds.df["feat_2"][2] - 0.2 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][0] - 1.4 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][1] - 5.9 < 0.00001)
    assert np.abs(phe.ds.df["feat_3"][2] - 0.34 < 0.00001)


def test_phe_getitem_setitem_delitem(dataset_iris):
    phe = Phenonaut(dataset_iris)
    phe["test_ds"] = dataset_iris.copy()
    assert len(phe.datasets) == 2
    assert len(phe[[0, 1]]) == 2
    assert phe["test_ds"].data.shape[1] == 4
    del phe["Iris"]
    assert len(phe.datasets) == 1
    assert phe.datasets[0].name == "test_ds"


def test_phe_save_and_load(dataset_iris):
    tmp_file = tempfile.NamedTemporaryFile(delete=True)
    try:
        phe = Phenonaut(dataset_iris)
        phe.save(tmp_file.name)
        phe_loaded = Phenonaut.load(tmp_file.name)
        assert len(phe_loaded.datasets) == 1
    finally:
        tmp_file.close()


def test_phe_clone(dataset_iris):
    phe = Phenonaut(dataset_iris)
    phe.clone_dataset("Iris", "Iris2")
    phe.clone_dataset("Iris", "Iris2", overwrite_existing=True)
    phe.clone_dataset("Iris", "Iris3", overwrite_existing=True)
    phe.clone_dataset("Iris", "Iris4", overwrite_existing=False)
    assert phe.keys() == ["Iris", "Iris2", "Iris3", "Iris4"]


def test_readin_transformations():
    df = pd.read_csv(
        StringIO(
            """"idx","Var1","Var2","value"
        
        "1","ENSG00000225972","A1_CPD1_PLATE1",4.0185
        "2","ENSG00000225630","A1_CPD1_PLATE1",1.1539
        "3","ENSG00000225972","A2_CPD2_PLATE1",10.6661
        "4","ENSG00000225630","A2_CPD2_PLATE1",1.6130
        "5","ENSG00000225972","A3_CPD3_PLATE1",0.1234
        "6","ENSG00000225630","A3_CPD3_PLATE1",9.8436
        "7","ENSG00000225972","A4_Pos_ctrl_PLATE1",0.1234
        "8","ENSG00000225630","A4_Pos_ctrl_PLATE1",9.8436"""
        )
    ).reset_index()
    phe = Phenonaut(
        df,
        "Test_df",
        metadata={
            "transforms": [
                ("replace_str", ("Var2", "Pos_", "Pos-")),
                ("split_column", ("Var2", "_", ["Well", "CpdID", "PlateID"])),
                ("pivot", ("Var1", "value")),
                #'transpose',
                # (pivot_table(values = 'value', index=["WellName"], columns = 'Var1')
            ],
            "features_prefix": "ENSG",
        },
    )
    print(phe.df)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from .generic_transformations import *
from .transformer import *
from .supervised_transformer import *
from .dimensionality_reduction import PCA, TSNE, UMAP
from .preparative import *



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from sklearn import decomposition
from phenonaut.data.dataset import Dataset
from typing import Callable, Type, Union, List
from inspect import isclass
from numpy import array
import numpy as np
import pandas as pd


class SupervisedTransformer:
    method = None
    method_kwargs = {}
    has_fit = False
    has_transform = False
    has_fit_transform = False
    new_feature_names = None
    is_callable = False
    callable_args = {}

    def __init__(self, method, new_feature_names=None, callable_args: dict = {}, **kwargs):

        # Here we need to handle a method being put in, a class, and a class instance.

        if isclass(method):
            method = method(**kwargs)
        if hasattr(method, "fit"):
            self.has_fit = True
        if hasattr(method, "transform"):
            self.has_transform = True
        if hasattr(method, "fit_transform"):
            self.has_fit_transform = True
        if callable(method):
            self.is_callable = True
            self.method_kwargs = kwargs
            self.callable_args = callable_args
        self.method = method
        self.new_feature_names = new_feature_names

    def _fit_df(
        self,
        X: Union[pd.DataFrame, np.array],
        y: Union[pd.DataFrame, np.array],
    ):
        if not self.has_fit:
            raise TypeError(f"Supplied method {self.method} has no fit function")
        self.method.fit(X, y)

    def _transform_df(
        self,
        df: pd.DataFrame,
    ) -> np.ndarray:

        if self.is_callable:
            transformed_data = self.method(df, **self.callable_args)
            if isinstance(transformed_data, pd.DataFrame):
                return transformed_data.values
            return transformed_data

        if not self.has_transform:
            raise TypeError(f"Supplied method {self.method} has no transform function")
        transformed_data = self.method.transform(df)
        if isinstance(transformed_data, pd.DataFrame):
            return transformed_data.values
        return transformed_data

    def fit(
        self,
        data: Union[Dataset, pd.DataFrame],
        y_or_ycolumnlabel: Union[str, pd.Series, np.array] = None,
        fit_perturbation_ids: Union[str, list] = None,
    ):

        # Dataframe
        if isinstance(data, pd.DataFrame):
            if isinstance(y_or_ycolumnlabel, str):
                if y_or_ycolumnlabel in data.columns:
                    X = data[list(data.columns).remove(y_or_ycolumnlabel)]
                    y = data[y_or_ycolumnlabel]
                    self._fit_df(X, y)
            else:
                self._fit_df(data, y_or_ycolumnlabel)
        else:
            if not isinstance(data, Dataset):
                raise TypeError(
                    f"fit requires a pd.DataFrame, or a Dataset class, received {type(data)}"
                )
            if fit_perturbation_ids is None:
                fit_perturbation_ids = list(data.get_unique_perturbations())
            if isinstance(y_or_ycolumnlabel, str):
                self._fit_df(
                    data.df.loc[
                        data.df.query(f"{data.perturbation_column} == @fit_perturbation_ids").index,
                        data.features,
                    ],
                    data.df.loc[y_or_ycolumnlabel],
                )
            else:
                self._fit_df(
                    data.df.loc[
                        data.df.query(f"{data.perturbation_column} == @fit_perturbation_ids").index,
                        data.features,
                    ],
                    y_or_ycolumnlabel,
                )

    def transform(
        self,
        data: Union[Dataset, pd.DataFrame],
        new_feature_names: List = None,
    ):

        # Do appropriate calls for DataFrame or Dataset to get transformed_data
        # which will be a np.array
        transformed_data = None
        if isinstance(data, pd.DataFrame):
            transformed_data = self.transform_df(data.df)
        else:
            if not isinstance(data, Dataset):
                raise TypeError(
                    f"transform requires a pd.DataFrame, or a Dataset class, received {type(data)}"
                )
        transformed_data = self._transform_df(data.df.loc[:, data.features])

        # Now we assign features/column names with the following descending priority:
        # new_feature_names (argument to this function)
        # self.new_feature_names
        # generated from self.method.__name__ with -n after, where x is the column number/feature count.
        if new_feature_names is None:
            new_feature_names = self.new_feature_names
        if new_feature_names is None:
            new_feature_names = [
                f"{self.method.__name__}-{n+1}" for n in range(transformed_data.shape[1])
            ]
        data.df[new_feature_names] = transformed_data
        data.features = (new_feature_names, f"{self.method}")

    def _fit_transform_df(
        self,
        df: pd.DataFrame,
    ) -> np.ndarray:

        if not self.has_fit_transform:
            raise TypeError(f"Supplied method {self.method} has no fit_transform function")

        transformed_data = self.method.fit_transform(df)
        if isinstance(transformed_data, pd.DataFrame):
            return transformed_data.values
        return transformed_data

    def fit_transform(
        self,
        data: Union[Dataset, pd.DataFrame],
        new_feature_names: List = None,
    ):

        # Do appropriate calls for DataFrame or Dataset to get transformed_data
        # which will be a np.array
        transformed_data = None
        if isinstance(data, pd.DataFrame):
            transformed_data = self._fit_transform_df(data.df)
        else:
            if not isinstance(data, Dataset):
                raise TypeError(
                    f"fit_transform requires a pd.DataFrame, or a Dataset class, received {type(data)}"
                )
        transformed_data = self._fit_transform_df(data.df.loc[:, data.features])

        # Now we assign features/column names with the following descending priority:
        # new_feature_names (argument to this function)
        # self.new_feature_names
        # generated from self.method.__name__ with -n after, where x is the column number/feature count.
        if new_feature_names is None:
            new_feature_names = self.new_feature_names
        if new_feature_names is None:
            new_feature_names = [
                f"{self.method.__name__}-{n+1}" for n in range(transformed_data.shape[1])
            ]
        data.df[new_feature_names] = transformed_data
        data.features = (new_feature_names, f"{self.method}")



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Optional, Union
import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F
from pathlib import Path
import numpy as np


class _DAVE_model(nn.Module):
    def _topol_to_inout_tuples(
        self, input: np.ndarray
    ) -> tuple[tuple[int, ...], tuple[int, int], tuple[tuple[int, ...]]]:
        """DAVE topology to view encoder sizes tuple list, embeding tuple, and decoder tuple list

        Given a view topology like [100, 77, 62, 32]
        We need to generate nn.Linear input and output size tuples for:
        view_encoder [(100, 77), (77, 62)]
        embedding (62, 32)
        view_decoder [(32, 62), (62, 77), (77, 100)]
        The above is returned as a 3 member tuple, for encoder,embedding and decoder, like:
        ([(100, 77), (77, 62)], (62, 32), [(32, 62), (62, 77), (77, 100)])

        Parameters
        ----------
        input : np.ndarray
            View encoder node sizes, like: ([100,  77,  62,  32])

        Returns
        -------
        tuple[tuple[int, ...], tuple[int, int], tuple[tuple[int, ...]]]
            3 member tuple containing:
            element 0 = List of encoder node input output tuples
            element 1 = 2 member tuple giving embedding space input and output sizes
            element 2 = List of decoder node input output tuples
        """
        encoding = [(input[0], input[1])]
        for i in range(2, len(input) - 1):
            encoding.append((encoding[-1][1], input[i]))
        embedding = (input[-2], input[-1])
        decoding = [(embedding[1], encoding[-1][1])] + list((t[1], t[0]) for t in encoding[::-1])
        return (encoding, embedding, decoding)

    def _assign_layers(self, topology: tuple[tuple[int, ...], tuple[int, ...]]):

        self.encoder1 = nn.ModuleList()
        self.encoder2 = nn.ModuleList()
        self.decoder1 = nn.ModuleList()
        self.decoder2 = nn.ModuleList()

        # Set up network nodes for view 1 - also do embedding here
        encoder_dims, embedding_dims, decoder_dims = self._topol_to_inout_tuples(topology[0])
        for dims in encoder_dims:
            self.encoder1.append(nn.Linear(*dims))
            self.encoder1.append(nn.ReLU())
        self.latent1 = nn.Linear(*embedding_dims)
        self.latent2 = nn.Linear(*embedding_dims)
        for dims in decoder_dims[:-1]:
            self.decoder1.append(nn.Linear(*dims))
            self.decoder1.append(nn.ReLU())
        self.decoder1.append(nn.Linear(*decoder_dims[-1]))
        self.decoder1.append(nn.Sigmoid())

        # Set up network nodes for view 2 - embedding already initialised above
        encoder_dims, embedding_dims, decoder_dims = self._topol_to_inout_tuples(topology[1])
        for dims in encoder_dims:
            self.encoder2.append(nn.Linear(*dims))
            self.encoder2.append(nn.ReLU())
        for dims in decoder_dims[:-1]:
            self.decoder2.append(nn.Linear(*dims))
            self.decoder2.append(nn.ReLU())
        self.decoder2.append(nn.Linear(*decoder_dims[-1]))
        self.decoder2.append(nn.Sigmoid())

    def __init__(
        self,
        view_sizes: tuple[int, int] = (100, 100),
        embedding_size: int = 32,
        n_hidden: Union[int, tuple[int, int], list[int, int]] = (3, 3),
        topology: tuple[tuple[int, ...], tuple[int, ...]] = None,
    ):
        super().__init__()
        if isinstance(n_hidden, int):
            n_hidden = (n_hidden, n_hidden)
        if topology is None:
            view_1 = np.linspace(view_sizes[0], embedding_size, n_hidden[0] + 2, dtype=int)
            view_2 = np.linspace(view_sizes[1], embedding_size, n_hidden[1] + 2, dtype=int)
            # Num inputs to latent space must be the same for each view.
            if view_1[-2] != view_2[-2]:
                if len(view_1) < 3 or len(view_2) < 3:
                    raise ValueError(
                        "Node sizes leading into embedding space were not the same, and with no hidden layers, "
                        "the sizes could not be averaged to give view 1 and view 2 pre-embedding layer the same number of inputs"
                    )
                pre_embedding_mean_layer_size = int((view_1[-2] + view_2[-2]) / 2)
                view_1[-2] = pre_embedding_mean_layer_size
                view_2[-2] = pre_embedding_mean_layer_size

            topology = [view_1, view_2]

        if topology[0][-1] != topology[1][-1]:
            raise ValueError(
                f"topology[0] and [1] did not both converge to the same number of embedding dimensions ({topology[0][-1]} and {topology[1][-1]})"
            )
        self.topology = topology
        self._assign_layers(topology)

    def encode_view1(self, x):
        for layer in self.encoder1:
            x = layer(x)
        return self.latent1(x), self.latent2(x)

    def encode_view2(self, x):
        for layer in self.encoder2:
            x = layer(x)
        return self.latent1(x), self.latent2(x)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def reparameterize_2_views(self, mu1, mu2, logvar1, logvar2):
        std1 = torch.exp(0.5 * logvar1)
        std2 = torch.exp(0.5 * logvar2)
        eps1 = torch.randn_like(std1)
        eps2 = torch.randn_like(std2)
        return mu1 + mu2 + eps1 * std1 + eps2 * std2

    def decode_view1(self, z):
        for layer in self.decoder1:
            z = layer(z)
        return z

    def decode_view2(self, z):
        for layer in self.decoder2:
            z = layer(z)
        return z

    def forwardA(self, x):
        z, _ = self.encode_view1(x)
        return self.decode_view1(z), self.decode_view2(z)

    def forwardB(self, x):
        z, _ = self.encode_view2(x)
        return self.decode_view1(z), self.decode_view2(z)

    def forward(self, xA, xB):
        muA, logvarA = self.encode_view1(xA)
        muB, logvarB = self.encode_view2(xB)
        z = self.reparameterize_2_views(muA, muB, logvarA, logvarB)
        return self.decode_view1(z), self.decode_view2(z), muA, muB, logvarA, logvarB


class DAVE:
    def __init__(
        self,
        batch_size=1024,
        learning_rate=1e-3,
        epochs=10,
        num_hidden_layers=1,
        seed: Optional[int] = None,
    ):
        if seed is not None:
            torch.manual_seed(seed)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.num_hidden_layers = num_hidden_layers
        self.model = _DAVE_model()

    # Reconstruction + KL divergence losses summed over all elements and batch
    def loss_function(self, recon_XA, recon_XB, XA, XB, muA, muB, logvarA, logvarB):
        bceA = F.binary_cross_entropy(recon_XA, XA, reduction="sum")
        bceB = F.binary_cross_entropy(recon_XB, XB, reduction="sum")
        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        kldA = -0.5 * torch.sum(1 + logvarA - muA.pow(2) - logvarA.exp())
        kldB = -0.5 * torch.sum(1 + logvarB - muB.pow(2) - logvarB.exp())
        return bceA + bceB + kldA + kldB

    def fit(self, X, y):
        if self.model.topology[0][0] != X.shape[1] or self.model.topology[1][0] != y.shape[1]:
            self.model = _DAVE_model(
                view_sizes=(X.shape[1], y.shape[1]), n_hidden=self.num_hidden_layers
            ).to(self.device)
        print(
            f"Constructed DAVE model on {self.device}, {self.batch_size=}, {self.learning_rate=}, {self.epochs=}, {self.num_hidden_layers=}"
        )
        y = torch.tensor(y.astype(np.float32))
        X = torch.tensor(X.astype(np.float32))
        train_tensor = torch.utils.data.TensorDataset(X, y)
        train_loader = torch.utils.data.DataLoader(
            dataset=train_tensor, batch_size=self.batch_size, shuffle=True
        )

        # optimizer = optim.SGD(self.model.parameters(), lr=learning_rate, momentum=0.9)
        optimizer = optim.Adam(self.model.parameters(), lr=self.learning_rate)

        criterion = nn.MSELoss()
        self.model.train()
        for epoch in range(self.epochs):  # loop over the dataset multiple times
            epoch_loss = 0
            for i, data in enumerate(train_loader):
                # get the inputs; data is a list of [inputs, labels]
                X_train, y_train = data
                X_train = X_train.to(self.device)
                y_train = y_train.to(self.device)

                # zero the parameter gradients
                optimizer.zero_grad()
                reconA, reconB, muA, muB, logvarA, logvarB = self.model(X_train, y_train)
                loss = self.loss_function(
                    reconA, reconB, X_train, y_train, muA, muB, logvarA, logvarB
                )
                loss.backward()
                epoch_loss += loss.item()
                optimizer.step()

    def predict(self, X):
        self.model.eval()
        with torch.no_grad():
            data = torch.tensor(X.astype(np.float32)).to(self.device)
            return self.model.decode_view2(self.model.encode_view1(data)[0]).to("cpu").numpy()



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Union
from pathlib import Path
import pandas as pd


def write_xlsx(
    output_file: Union[Path, str],
    dfs: Union[pd.DataFrame, list[pd.DataFrame], dict[str, pd.DataFrame]],
    df_names: Union[tuple, str] = None,
):
    """Write an XLSX spreadsheet file

    Writing an XLSX file allows multiple sheets to be embedded in a single file
    which may be advantageous over CSV/TSV in many situations.

    Parameters
    ----------
    output_file : Union[Path, str]
        Output file path.
    dfs : Union[pd.DataFrame, list[pd.DataFrame], dict[str, pd.DataFrame]]
        DataFrame for output, list of DataFrames, or dictionary of DataFrames.
        If a single DataFrame is given, or list of DataFrames, then the df_names
        argument must contain a str, or list of str respectively. If a
        dictionary is passed for this parameter, then the keys under which each
        DataFrame exist are used as sheet names.
    df_names : Union[tuple, str], optional
        The name of sheets under which DataFrames are stored in the spreadsheet.
        If a dictionary was given for the dfs parameter, then this parameter has
        no effect. If this parameter is None, then sheets are named 'Sheet#N',
        where #N is a number starting at 1 and incremented with each addition of
        a new sheet. By default None.

    Raises
    ------
    ValueError
        output_file should be Path or str
    """
    if isinstance(output_file, str):
        output_file = Path(output_file)
    if isinstance(output_file, Path):
        output_file = output_file.resolve()
        if not output_file.parent.exists():
            output_file.parent.mkdir(parents=True)
    else:
        raise ValueError(f"output_file should have been a Path or str, it was {type(output_file)}")

    if isinstance(dfs, dict):
        df_names = list(dfs.keys())
        dfs = list(dfs.values())

    if isinstance(dfs, pd.DataFrame):
        dfs = [dfs]

    if df_names is None:
        df_names = [f"Sheet{i+1}" for i in range(len(dfs))]
    if isinstance(df_names, str):
        df_names = [df_names]
    if len(dfs) > len(df_names):
        num_to_add = len(dfs) - len(df_names)
        df_names.extend([f"Sheet{i+1}" for i in range(len(df_names), len(df_names) + num_to_add)])

    with pd.ExcelWriter(output_file) as writer:
        for df, sheet_name in zip(dfs, df_names):
            df.to_excel(writer, sheet_name=sheet_name)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Optional, Union
from phenonaut.data import Dataset
from phenonaut import Phenonaut
import numpy as np
import pandas as pd
from scipy.linalg import svd
from sklearn.preprocessing import StandardScaler as SK_StandardScaler

from sklearn.linear_model import LinearRegression as SklearnLinreg
from phenonaut.transforms.transformer import Transformer

"""This module contains functionality to remove highly correlated features from phenonaut datasets"""


class RemoveHighlyCorrelated:
    def __init__(self, verbose: bool = False):
        """RemoveHighlyCorrelated

        This class calculates correlations between all features and allows either removal
        of features correlated above a given theshold, or uses the iterative
        removal of features with the highest R^2 against another feature.

        Initialisation takes no arguments. Once constructed, the object can be called
        directly which is the same as a call to the filter method.

        Parameters:
        ----------
        verbose : bool, optional
            If True, then details of the removed features are printed.
        """
        self.verbose = verbose

    def __call__(
        self,
        ds: Dataset,
        threshold: Union[float, None] = 0.90,
        min_features: Optional[int] = None,
        drop_columns: bool = False,
        **corr_kwargs,
    ):
        """Run RemoveHighlyCorrelated - identical to calling the filter function.

        Parameters
        ----------
        ds : Dataset
            The dataset from which features are to be removed
        threshold : Union[float, None]
            The threshold value for calculated correlations, above which, a feature should
            be removed. If None, then it is expected that the n argument is given and features
            with the highest correlation are then iteratively removed. By default 0.9
        min_features : int, optional
            The number of features to keep. If the threshold argument is None, then features
            are iteratively removed, ordered by the most correlated until the number of features
            is equal to n. If threshold is not None, and is a float, then n acts as a minimum
            number of features and feature removal will stop, no matter the correlations present
            in the dataset.
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        corr_kwargs : dict
            Keyword arguments which may be passed to the pd.Dataframe.corr function, allowing
            chaniging of the correlation calculation method from pearson to otheres.
        """
        self.filter(
            ds,
            threshold=threshold,
            min_features=min_features,
            drop_columns=drop_columns,
            **corr_kwargs,
        )

    def filter(
        self,
        ds: Dataset,
        threshold: Union[float, None] = 0.9,
        min_features: Optional[int] = None,
        drop_columns: bool = False,
        **corr_kwargs,
    ):
        """Run the RemoveHighlyCorrelated filter

        Parameters
        ----------
        ds : Dataset
            The dataset from which features are to be removed
        threshold : Union[float, None]
            The threshold value for calculated correlations, above which, a features should
            be removed. If None, then it is expected that the n argument is given and features
            with the highest correlation are then iteratively removed. Note that the absolute
            value of the calculated correlation coefficients are taken, so this number should
            always be positive, by default 0.9.
        min_features : int, optional
            The number of features to keep. If the threshold argument is None, then features
            are iteratively removed, ordered by the most correlated until the number of features
            is equal to n. If threshold is not None, and is a float, then n acts as a minimum
            number of features and feature removal will stop, no matter the correlations present
            in the dataset.
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        corr_kwargs : dict
            Keyword arguments which may be passed to the pd.Dataframe.corr function, allowing
            chaniging of the correlation calculation method from pearson to otheres.
        """
        if threshold is not None:
            self._filter_threshold(ds, threshold, drop_columns=drop_columns, **corr_kwargs)
        if min_features is not None:
            self._filter_leaving_n(ds, min_features, drop_columns=drop_columns, **corr_kwargs)

    def _filter_leaving_n(self, ds: Dataset, n: int, drop_columns: bool = False, **corr_kwargs):
        """Filter the dataset, iteratively removing highest correlated features

        Parameters
        ----------
        ds : Dataset
            The dataset to be filtered.
        n : int
            The feature count to stop at.
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        """
        removed_features = []
        removed_features_correlations = []
        ds_features = ds.features
        while len(ds_features) > n:
            corr_matrix = np.triu(ds.df[ds_features].corr(**corr_kwargs).abs(), k=1)
            max_correlations = np.max(corr_matrix, axis=0)
            argmax = np.argmax(max_correlations)
            removed_features.append(ds_features[argmax])
            removed_features_correlations.append(max_correlations[argmax])
            ds_features.remove(removed_features[-1])
            if self.verbose:
                print(
                    f"Removed {removed_features[-1]} feature, corr={max_correlations[argmax]}, {len(ds_features)} features remaining"
                )
        ds.features = (
            ds_features,
            f"Removed {len(removed_features)} highly correlated features '{removed_features}', with correlation coefs of {removed_features_correlations}, attempting to reduce to {n} most uncorrelated features",
        )
        if drop_columns:
            ds.drop_columns(
                removed_features,
                " because columns were old features removed by RemoveHighlyCorrelated",
            )

    def _filter_threshold(
        self, ds: Dataset, threshold: float = 0.9, drop_columns: bool = False, **corr_kwargs
    ):
        """Filter dataset by feature correlation threshold

        Parameters
        ----------
        ds : Dataset
            The dataset to be filtered
        threshold : float, optional
            Threshold correlation value. In the case of pearson correlation coefficient etc,
            then 1 indicates perfect correlation. Note that the absolute value of the calculated
            correlation coefficients are taken, so this number should always be positive,
            by default 0.9
        corr_kwargs : dict
            Keyword arguments which may be passed to the pd.Dataframe.corr function, allowing
            chaniging of the correlation calculation method from pearson to otheres.
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        """
        removed_features = set()  # If we choose to delete a column, then place it here.
        corr_matrix = ds.data.corr(**corr_kwargs).abs()
        for i in range(len(corr_matrix.columns)):
            if np.max(corr_matrix.iloc[i, :i]) >= threshold:
                removed_features.add(corr_matrix.columns[i])
                if self.verbose:
                    print(
                        f"Removed {corr_matrix.columns[i]} feature, corr={np.max(corr_matrix.iloc[i, :i])}, {len(removed_features)-len(ds.features)} features remaining"
                    )
        uncorrelated_features = [f for f in ds.features if f not in removed_features]
        ds.features = (
            uncorrelated_features,
            f"Removed {len(removed_features)} features with correlations>='{threshold}'",
        )
        if drop_columns:
            ds.drop_columns(
                removed_features,
                " because columns were old features removed by RemoveHighlyCorrelated",
            )


class VIF:
    def __init__(self, verbose: bool = False):
        """VIF - Variance Inflation Factor filter

        This class calculates the VIF (Variance Inflation Factor) for features within
        a dataset. This can be a computationally expensive process as the number of
        linear regressions required to be run is almost N^2 with features.

        Initialisation takes no arguments. Once constructed, the object can be called
        directly which is the same as a call to the filter method.

        Parameters:
        ----------
        verbose : bool, optional
            If True, then details of the removed features are printed.
        """
        self.verbose = verbose

    def __call__(self, ds, vif_cutoff: float = 5.0, min_features=2, drop_columns: bool = False):
        """Run the variance inflation factor filter, the same as calling the filter method.

        Parameters
        ----------
        ds : Dataset
            The phenonaut dataset to be operated on
        vif_cutoff : float, optional
            The variance inflation factor cutoff.  Values above 5.0 indicate strong
            correlation, by default 5.0
        min_features : int, optional
            Remove features by VIF score all the way down to a minimum given by this
            argument, by default 2
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        """
        self.filter(ds, vif_cutoff=5.0, min_features=2, drop_columns=False)

    def get_vif_scores(
        self, ds: Dataset, use_features: Optional[list[str]] = None
    ) -> dict[str, float]:
        """Get VIF scores dictionary from a phenonaut Dataset

        Parameters
        ----------
        ds : Dataset
            The dataset from which to calculate VIF scores
        use_features : Optional[list[str]], optional
            If None, then the features found within the dataset with a call to Dataset.features
            are used. This behaviour can be changed by passing a list of features as this
            use_features argument, by default None

        Returns
        -------
        dict[str, float]
            Dictionary of mapping feature name to VIF score
        """
        if use_features is None:
            use_features = ds.features
        vif_scores = {}
        for held_out_feat in use_features:
            X = ds.df[[feat for feat in use_features if feat != held_out_feat]].values
            y = ds.df[held_out_feat].values
            vif_scores[held_out_feat] = 1 / (1 - SklearnLinreg().fit(X, y).score(X, y))
        return vif_scores

    def filter(self, ds, vif_cutoff: float = 5.0, min_features=2, drop_columns: bool = False):
        """Run the variance inflation factor filter

        Parameters
        ----------
        ds : Dataset
            The phenonaut dataset to be operated on
        vif_cutoff : float, optional
            The variance inflation factor cutoff.  Values above 5.0 indicate strong
            correlation, by default 5.0
        min_features : int, optional
            Remove features by VIF score all the way down to a minimum given by this
            argument, by default 2
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        """
        vif_scores = self.get_vif_scores(ds)
        ds_features = ds.features
        removed_features = []
        removed_features_scores = []
        while max(vif_scores.values()) > vif_cutoff and len(ds_features) > min_features:
            feature_with_highest_vif_score = max(vif_scores, key=vif_scores.get)
            removed_features.append(feature_with_highest_vif_score)
            removed_features_scores.append(max(vif_scores.values()))
            ds_features.remove(feature_with_highest_vif_score)
            vif_scores = self.get_vif_scores(ds, use_features=ds_features)
            if self.verbose:
                print(
                    f"Removed {feature_with_highest_vif_score} feature, VIF={removed_features_scores[-1]}, {len(ds_features)} features remaining"
                )

        if drop_columns:
            ds.drop_columns(
                removed_features, reason=f"VIF filtered, {vif_cutoff=}, {min_features=}"
            )
        else:
            ds.features = (
                ds_features,
                f"VIF filtering removed '{removed_features}' feature, {vif_cutoff=}, {min_features=}",
            )


class RemoveHighestCorrelatedThenVIF:
    def __init__(self, verbose: bool = False):
        """RemoveHighestCorrelatedThenVIF

        Ideally, VIF would be applied to very large datasets.  Due to the almost n^2 number
        of linear regression required as features increase, this is not possible on datasets
        with a large number of features - such as methylation datasets.  We therefore must
        use other methods to reduce the features to a comfortable level allowing VIF to be
        performed. This class calculates correlations between all features and iteratively
        removes the features with the highest R^2 against another feature. Once the number
        of featurs is reduced to a level suitable for VIF, VIF is performed.

        Initialisation takes no arguments. Once constructed, the object can be called
        directly which is the same as a call to the filter method.

        Parameters:
        ----------
        verbose : bool, optional
            If True, then details of the removed features are printed.
        """
        self.verbose = verbose

    def __call__(
        self,
        ds: Dataset,
        n_before_vif=1000,
        vif_cutoff: float = 5.0,
        min_features=2,
        drop_columns=False,
        **corr_kwargs,
    ):
        """Run RemoveHighestCorrelatedThenVIF

        Parameters
        ----------
        ds : Dataset
            The dataset from which features are to be removed
        n_before_vif : int, optional
            The number of features to remove before applying VIF. This is required when
            dealing with large datasets which would be too time consuming to process entirely
            with VIF.  Features are removed iteratively, selecting the most correlated
            features and removing them, by default 1000
        vif_cutoff : float, optional
            The VIF cutoff value, above which features are removed. Features with VIF scores
            above 5.0 are considered highly correlated, by default 5.0.
        min_features : int, optional
            Remove features by VIF score all the way down to a minimum given by this
            argument, by default 2
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        corr_kwargs : dict
            Keyword arguments which may be passed to the pd.Dataframe.corr function, allowing
            chaniging of the correlation calculation method from pearson to otheres.
        """
        self.filter(
            ds,
            n_before_vif=n_before_vif,
            vif_cutoff=vif_cutoff,
            drop_columns=drop_columns,
            **corr_kwargs,
        )

    def filter(
        self,
        ds: Dataset,
        n_before_vif=1000,
        vif_cutoff: float = 5.0,
        min_features=2,
        drop_columns=False,
        **corr_kwargs,
    ):
        """Run RemoveHighestCorrelatedThenVIF

        Parameters
        ----------
        ds : Dataset
            The dataset from which features are to be removed
        n_before_vif : int, optional
            The number of features to remove before applying VIF. This is required when
            dealing with large datasets which would be too time consuming to process entirely
            with VIF.  Features are removed iteratively, selecting the most correlated
            features and removing them, by default 1000
        vif_cutoff : float, optional
            The VIF cutoff value, above which features are removed. Features with VIF scores
            above 5.0 are considered highly correlated, by default 5.0.
        min_features : int, optional
            Remove features by VIF score all the way down to a minimum given by this
            argument, by default 2
        drop_columns : bool, optional
            If drop columns is True, then not only will features be removed from the
            dataset features list, but the columns for these features will be removed
            from the dataframe, by default False
        corr_kwargs : dict
            Keyword arguments which may be passed to the pd.Dataframe.corr function, allowing
            chaniging of the correlation calculation method from pearson to otheres.
        """
        rhc = RemoveHighlyCorrelated(self.verbose)
        rhc(ds, threshold=None, min_features=n_before_vif, drop_columns=drop_columns, **corr_kwargs)
        vif = VIF(self.verbose)
        vif(ds, drop_columns=drop_columns, min_features=min_features)


class __ZCA__:
    """
    Private class used by ZCA, implementing fit, transform and
    inverse_transform, heavily inspired by https://github.com/mwv/zca/
    """

    def __init__(self, eps=1e-6):
        self.eps = eps

    def fit(self, X: np.ndarray):
        self.mean = np.mean(X, axis=0)
        centered_X = X - self.mean
        U, S, _ = svd(np.dot(centered_X.T, centered_X) / (centered_X.shape[0] - 1))
        s = np.sqrt(np.clip(S, self.eps, None))
        self.W = np.dot(np.dot(U, np.diag(1.0 / s)), U.T)
        self.deW = np.dot(np.dot(U, np.diag(s)), U.T)
        return self

    def transform(self, X):
        return np.dot(X - self.mean, self.W.T)

    def inverse_transform(self, X):
        return np.dot(X, self.deW) + self.mean


class ZCA(Transformer):
    """ZCA whitening

    Can be instantiated and called like:

    .. code-block:: python

        zca=ZCA()
        zca(dataset)

    Parameters
    ----------
    dataset : Dataset
        The Phenonaut dataset on which to apply ZCA whitening
    new_feature_names : Union[list[str], str]
        List of strings containing the names for the new features. Can also
        be just a single string, which then has numerical suffixes attached
        enumerating the number of new features generated. By default "ZCA"
    """

    def __init__(self, new_feature_names: Union[list[str], str] = "ZCA_"):
        super().__init__(__ZCA__, new_feature_names=new_feature_names, transformer_name="ZCA")


class __RobustMAD__:
    def __init__(self, mad_scale: Union[str, float] = "normal", epsilon: Optional[float] = None):
        """RobustMAD

        Robust scaling, as used popularly in pycytominer

        https://github.com/cytomining/pycytominer/blob/master/pycytominer/operations/transform.py

        Perform robust normalization using median and mad
        (median absolute deviation), like:

        X=X-median/(mad+epsilon)

        May be instantiated and then called, or fit and then transform performed.
        Optional query argument carries out a pandas query to select relevant rows
        from the dataframe to perform the fit.

        Parameters
        ----------
        mad_scale : Union[str,float], optional
            Scaling argument passed to the median_abs_deviation
            scipy function. The value 'normal' implies the data
            is normally distributed, but this may be changed
            for other distribution types. See scipy documentation
            for the function for a more complete explanation.
            By default 'normal'.
        epsilon : float, optional
            Small fudge-factor to remove danger of median value.
            If None, then it is set by numpy eps value.
            By default None.
        """
        self._epsilon = np.finfo(np.float64).eps if epsilon is None else epsilon
        self._mad_scale = mad_scale

    def fit(self, data: np.ndarray):
        """Fit the RobustMAD scaler to a dataset

        Parameters
        ----------
        dataset : Dataset
            Dataset upon which the fit is to be performed
        query : Optional[str], optional
            Pandas query which may be used to select rows, by default None

        Returns
        -------
        RobustMAD
            Returns a reference to itself as an object.
        """
        from scipy.stats import median_abs_deviation

        self._median = np.median(data, axis=0)
        self._mad = median_abs_deviation(data, nan_policy="omit", scale=self._mad_scale, axis=0)
        return self

    def transform(self, dataset: np.ndarray):
        return (dataset - self._median) / (self._mad + self._epsilon)


class RobustMAD(Transformer):
    def __init__(
        self,
        mad_scale: Union[str, float] = "normal",
        epsilon: Optional[float] = None,
        new_feature_names: Union[list[str], str] = "RobustMAD_",
    ):
        """RobustMAD

        Robust scaling, as used popularly in pycytominer

        https://github.com/cytomining/pycytominer/blob/master/pycytominer/operations/transform.py

        Perform robust normalization using median and mad
        (median absolute deviation), like:

        X=X-median/(mad+epsilon)

        May be instantiated and then called, or fit and then transform performed.
        Optional query argument carries out a pandas query to select relevant rows
        from the dataframe to perform the fit.

        Parameters
        ----------
        mad_scale : Union[str,float], optional
            Scaling argument passed to the median_abs_deviation
            scipy function. The value 'normal' implies the data
            is normally distributed, but this may be changed
            for other distribution types. See scipy documentation
            for the function for a more complete explanation.
            By default 'normal'.
        epsilon : float, optional
            Small fudge-factor to remove danger of median value.
            If None, then it is set by numpy eps value.
            By default None.
        new_feature_names : Union[list[str], str]
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. By default "RMAD"
        """

        self._epsilon = np.finfo(np.float64).eps if epsilon is None else epsilon
        self._mad_scale = mad_scale
        super().__init__(
            __RobustMAD__,
            new_feature_names,
            transformer_name="RMAD",
            constructor_kwargs={"mad_scale": mad_scale, "epsilon": epsilon},
        )


class StandardScaler(Transformer):
    def __init__(self, new_feature_names="StdScaler_"):
        """StandardScaler

        Perform StandardScaler on a dataset

        Can be instantiated and called like:

        .. code-block:: python

        scaler=StandardScaler()
        scaler(dataset)
        # or
        scaler(phe)
        # or
        scaler.fit_transform(dataset)
        # or
        scaler.fit(dataset)
        scaler.transform(dataset)


        """
        super().__init__(
            SK_StandardScaler,
            new_feature_names=new_feature_names,
            transformer_name="StandardScaler",
        )



# Copyright © The University of Edinburgh, 2023.
# Development has been supported by GSK.

from .distances import *
from .measures import scalar_projection, feature_correlation_to_target
from .performance import percent_replicating



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from .scatter import Scatter
from .visualisation_base import PhenonautVisualisation
from .spreadsheet import write_xlsx
from .heatmap import write_heatmap_from_df



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from pathlib import Path
import phenonaut
from pandas.errors import DataError
from typing import Union
import re
from phenonaut.transforms.preparative import RemoveHighlyCorrelated
from phenonaut.utils import load_dict


class Workflow:
    """Phenonaut Workflows allow operation through simple YAML workflows.

    Workflows may be defined in Phenonaut, and the module executed directly,
    rather than imported and used by a Python program. Workflows are
    defined using the simple YAML file format. However, due to the way in
    which they are read in, JSON files may also be used. As YAML files can
    contain multiple YAML entries, we build on this concept, allowing
    multiple workflows to be defined in a single YAML (or JSON) file. Once
    read in, workflows are dictionaries. From Python 3.6 onwards,
    dictionaries are ordered. We can therefore define our workflows in order
    and guarantee that they will be executed in the defined order. A
    dictionary defining workflows have the following structure:
    {job_name: task_list}, where job_name is a string, and task list is a
    list defining callable functions, or tasks required to complete the job.

    The job list takes the form of a list of dictionaries, each containning
    only one key which is the name of the task to be peformed. The value
    indexed by this key is a dictionary of argument:value pairs to be passed
    to the function responsible for performing the task. The structure is
    best understood with an example. Here, we see a simple workflow
    contained within a YAML file for calculation of the scalar projection
    phenotypic metric. YAML files start with 3 dashes.

    .. code-block:: console

        ---
        scalar_projection_example:
        - load:
            file: screening_data.csv
            metadata:
                features_prefix:
                    - feat_
        - scalar_projection:
            target_treatment_column_name: control
            target_treatment_column_value: pos
            output_column_label: target_phenotype
        - write_multiple_csvs:
            split_by_column: PlateID
            output_dir: scalar_projection_output


    The equivalent JSON with clearer (for Python programmers) formatting for the above is:

    .. code-block:: python

        {
            "scalar_projection_example": [
                {
                    "load": {
                    "file": "screening_data.csv",
                    "metadata": {
                    "features_prefix": ["feat_"]}
                    }
                },
                {
                    "scalar_projection": {
                        "target_treatment_column_name": "control",
                        "target_treatment_column_value": "pos",
                        "output_column_label": "target_phenotype",
                    }
                },
                {"write_multiple_csvs":{
                    "split_by_column": "PlateID",
                    "output_dir": "scalar_projection_output/"
                    }
                },
            ]
        }


    The workflow define above in the example YAML and JSON formats, has the
    name "scalar_projection_example", and consists of 3 commands.

    #. load
    #. scalar_projection
    #. write_multiple_csvs

    See the user guide for a full listing of commands.

    Parameters
    ----------
    workflow_path : Union[Path, str, dict]
        Workflows can be defined in YML or JSON files with their locations
        supplied or jobs passed as dictionaries. Dictionary keys denote the
        job names. Values under these keys should be lists of dictionaries.
        Each dictionary should have one key, denoting the name of the task
        and values under this key contain options for the called functions/
        tasks.

    Raises
    ------
    TypeError
        Supplied Path or str to file location does not appear to be a YAML or JSON
        file.
    """

    def __init__(self, workflow_path: Union[Path, str, dict]):
        self.workflow = load_dict(workflow_path)
        self.run_workflow()

    def run_workflow(self):
        """Run the workflow defined in the workflow object instance"""

        if self.workflow is None:
            raise ValueError("Workflow not loaded")
        # Iterate jobs defined in the YAML file. Typically only one job will be defined,
        # but multiple are allowed.
        for job in self.workflow:
            self.phe = phenonaut.Phenonaut(name=job)
            # Jobs are built by
            for task in self.workflow[job]:
                if isinstance(task, str):
                    task = {task: {}}
                if len(task.keys()) > 1:
                    raise KeyError(
                        "Phenonaut workflow tasks require one top level dictionary entry - like 'load' or 'filter_columns'"
                    )
                command = next(iter(task))
                arguments = task[command]
                print(f"Performing : {command}")
                try:
                    getattr(self, command)(arguments)
                except AttributeError as error:
                    raise AttributeError(f"{command} not implemented in Phenonaut Workflow")

    def load(self, arguments: dict):
        """Workflow function: load a dataset (CSV or PakcagedDataset)

        Workflow runnable function allowing loading of dataset from CSV or a
        PackagedDataset. As with all workflow runnable functions, this is
        designed to be called from a worflow.

        There are 2 possible options for loading in a dataset.

        Firstly, loading a user supplied CSV file.
        This option is initiated through inclusion of the 'file' key within
        arguments. The value under the 'file' key should be a string or Path to
        the CSV file.  In addition, a 'metadata' key is also required to be
        present in arguments, with a dictionary as the value. Within this
        dictionary under 'metadata', special keywords allow the reading of data
        in different formats. Special keys for the metadata dictionary are
        listed below (See Pandas documentation for read_csv for more in-depth
        information):

        sep
            define separator for fields within the file (by default ',')

        skiprows
            define a number of rows to skip at the beginning of a file

        header_row_number
            may be a single number, or list of numbers denoting header rows.

        transpose
            In the case of some transcriptomics raw data,
            transposition is required to have samples row-wise. Therefore the
            table must be transposed.  Set to True to transpose.

        index_col
            Column to use as the row labels of the CSV given as either
            the column names or as numerical indexes. Note: index_col=False can
            be used to force pandas to not use the first column as the index,
            e.g. when you have a malformed file with delimiters at the end of
            each line.

        key
            If an h5 file is supplied, then this is the key to the underlying pandas
            dataframe.

        Secondly, we may load a packaged dataset, by including the 'dataset' key
        within the arguments dictionary. The value under this key should be one
        of the current packaged datasets supported by workflow mode - currently
        TCGA, CMAP, Iris, Iris_2_views, BreastCancer. An additional key in the
        dictionary can be 'working_dir' with the value being a string denoting
        the location of the file data on a local filesystem, or the location
        that it should be downloaded to and stored.

        Parameters
        ----------
        arguments : dict
            Dictionary containing file and metadata keys, see function
            description/help/docstring.
        """
        if "file" in arguments and "dataset" in arguments:
            raise KeyError(
                "Cannot have both file and dataset arguments present within a load command"
            )

        if "file" in arguments:
            if "metadata" not in arguments:
                message = "Workflow load called without appropriate metadata arguments"
                raise KeyError(message)
            file = arguments["file"]
            metadata = arguments["metadata"]
            if file.endswith(".tsv") or file.endswith(".tsv.gz"):
                if "sep" not in metadata:
                    metadata["sep"] = "\t"
            if file.endswith(".h5"):
                if "key" not in arguments:
                    raise KeyError(
                        "Workflow load called on an h5 file, without supplying a key argument"
                    )
                self.phe.load_dataset(
                    "Workflow loaded dataset", file, metadata=metadata, h5_key=arguments["key"]
                )
            self.phe.load_dataset("Workflow loaded dataset", file, metadata=metadata)
            return
        if "dataset" in arguments:
            packaged_dataset_name = metadata["dataset"]
            working_dir = arguments.get("working_dir", None)
            if packaged_dataset_name == "TCGA":
                from phenonaut.packaged_datasets import TCGA

                self.phe = phenonaut.Phenonaut(TCGA(working_dir))
                return
            if packaged_dataset_name == "CMAP":
                from phenonaut.packaged_datasets import CMAP

                self.phe = phenonaut.Phenonaut(CMAP(working_dir))
                return
            if packaged_dataset_name == "Iris":
                from phenonaut.packaged_datasets import Iris

                self.phe = phenonaut.Phenonaut(Iris(working_dir))
                return
            if packaged_dataset_name == "Iris_2_views":
                from phenonaut.packaged_datasets import Iris_2_views

                self.phe = phenonaut.Phenonaut(Iris_2_views(working_dir))
                return
            if packaged_dataset_name == "BreastCancer":
                from phenonaut.packaged_datasets import BreastCancer

                self.phe = phenonaut.Phenonaut(BreastCancer(working_dir))
                return
            raise KeyError("Name of packaged dataset not found or accessible within a workflow")
        raise KeyError("Did not find file or dataset argument for load command")

    def filter_rows(self, arguments: dict):
        """Workflow function: Filter rows

        Designed to be called from a workflow, filter_rows alows keeping of only
        rows with a certain value in a certain column. Takes as arguments a
        dictionary containing query_column key:value pair and one of
        query_value, query_values or values key:value pairs:

            query_column
                name of the column that should match the value below

            query_value
                value to match

            query_values
                values to match (as a list)

            values
                values to match (as a list)

        Additionally, a key "keep" with a boolean value may be included. If True
        then rows matching the query are kept, if False, then rows matching are
        discarded, and non-matching rows kept.

        Parameters
        ----------
        arguments : dict
            Dictionary containing query_column key and value defining the column
            name, and one of the following keys: query_value, query_values,
            values. If plural, then values under the key should be a list
            containing  values to perform matching on, otherwise, singular
            value.

        Raises
        ------
        DataError
            [description]
        """
        values = []
        for key in ["query_value", "query_values", "values"]:
            if key in arguments:
                if isinstance(arguments[key], list):
                    values.extend(arguments[key])
                else:
                    values.append(arguments[key])
        if values == []:
            raise DataError("No query_value or query_values found to filter on")
        keep = arguments["keep"]
        query_column = arguments["query_column"]
        ## filter here
        self.phe.ds.filter_rows(query_column, values, keep=keep)

    def filter_columns(self, arguments: dict):
        """Workflow function: filter columns

        Designed to be called from a workflow, Datasets may have columns
        defined for keeping or removal. This function also provides a convenient
        way to reorder dataframe columns.

        Parameters
        ----------
        arguments : dict
            Dictionary of options, can include the following keys:
                keep: bool, optional, by default True
                    Only matching columns are kept if true.  If false, they are
                    removed.
                column_names: [list, str]
                    List of columns to keep (or regular expressions to match)
                column_name: str
                    Singular column to keep (or regular expressions to match)
                regex: bool, optional, by default False.
                    perform regular expression matching

        """
        keep = arguments.pop("keep", None)
        use_regex = arguments.pop("regex", False)
        if keep is None:
            message = "filter_columns in a workflow must have a keep argument, taking values True (to keep these columns), or False (to remove the columns)"
            raise KeyError(message)

        values = []
        for key in [
            "column_name",
            "column_names",
            "values",
            "column_value",
            "column_values",
        ]:
            if key in arguments:
                if isinstance(arguments[key], list):
                    values.extend(arguments[key])
                else:
                    values.append(arguments[key])
        if values == []:
            raise DataError("No query_value or query_values found to filter on")
        ## filter here
        self.phe.ds.filter_columns(values, keep=keep, regex=use_regex)
        return

    def write_multiple_csvs(self, arguments: dict):
        """Workflow function: Write multiple CSV files

        Designed to be called from a workflow. Often it is useful to write a CSV
        file per plate within the dataset, or group the data by some other
        identifier.

        Parameters
        ----------
        arguments : dict
            Dictionary, should contain:

            split_by_column: str
                the column to be split on

            output_dir: str
                the target output directory

            file_prefix: str
                optional prefix for each file.

            file suffix: str
                optional suffix for each file.

            file_extension: str
                optional file extension, by default '.csv'

        """
        split_by_column = arguments.pop("split_by_column")
        output_dir = arguments.pop("output_dir", Path.cwd())
        file_prefix = arguments.pop("file_prefix", "")
        file_suffix = arguments.pop("file_suffix", "")
        file_extension = arguments.pop("file_extension", ".csv")
        self.phe.ds.df_to_multiple_csvs(
            split_by_column,
            output_dir=output_dir,
            file_prefix=file_prefix,
            file_suffix=file_suffix,
            file_extension=file_extension,
            **arguments,
        )

    def write_csv(self, arguments: dict):
        """Workflow function: Write dataframe to CSV file.

        Designed to be called from a workflow, writes a CSV file using the
        Pandas.DataFrame.to_csv function.  Expects a dictionary as arguments
        containing a 'path' key, with a string value pointing at the destination
        location of the CSV file. Additional keys within the supplied dictionary
        are supplied to Pandas.DataFrame.to_csv as kwargs, allowing fully
        flexible output.

        Parameters
        ----------
        arguments : dict
            Dictionary should contain a 'path' key, may also contain a target_dataset key
            which if absent, defaults to -1 (usually the last added dataset).
        """
        file_path = arguments.pop("path")
        self.phe[arguments.get("target_dataset", -1)].df_to_csv(file_path, **arguments)

    def scalar_projection(self, arguments: dict):
        """Workflow function: Add a column for the scalar projection to a target perturbation.

        Designed to be called from a workflow, calculates the scalar projection
        and scalar rejection, quantifying on and off target phenotypes, as
        used in:
        Heiser, Katie, et al. "Identification of potential treatments for
        COVID-19 through artificial intelligence-enabled phenomic analysis
        of human cells infected with SARS-CoV-2." BioRxiv (2020).

        Parameters
        ----------
        arguments : dict, should contain:
            target_dataset
                Index or name of dataset which should be used in the measurment.
                If absent, then behaviour is as if -1 is supplied, indicating
                the last added dataset.
            target_perturbation_column_name
                normally a 'control' column
            target_perturbation_column_value:
                value to be found in the column defined previously.
            output_column_label:
                Output from the scalar projection will have the form: on_target_<output_column_label> and off_target_<output_column_label>
                if this is missing, then it is set to target_perturbation_column_value

        """
        if "output_column_label" not in arguments:
            arguments["output_column_label"] = arguments["target_perturbation_column_value"]
        phenonaut.metrics.scalar_projection(
            self.phe[arguments.get("target_dataset", -1)], **arguments
        )

    def euclidean_distance(self, arguments: dict):
        """Workflow function: Add a column for the euclidean distance to a target perturbation.

        Designed to be called from a workflow, calculates the euclidean distance
        in feature space.

        Parameters
        ----------
        arguments : dict, should contain:
            target_dataset
                Index or name of dataset which should be used in the measurment.
                If absent, then behaviour is as if -1 is supplied, indicating
                the last added dataset.
            target_perturbation_column_name
                normally a 'control' column
            target_perturbation_column_value:
                value to be found in the column defined previously.
            output_column_label:
                Output column for the measurement. If this is missing, then it is set to target_perturbation_column_value.

        """
        if "output_column_label" not in arguments:
            arguments["output_column_label"] = arguments["target_perturbation_column_value"]
        target_perturbation_column_name = arguments["target_perturbation_column_name"]
        target_perturbation_column_value = arguments["target_perturbation_column_value"]
        output_column_label = arguments["output_column_label"]
        ds = self.phe[arguments.get("target_dataset", -1)]
        ds.df[output_column_label] = phenonaut.metrics.euclidean(
            ds.data,
            ds.df.query(
                f"{target_perturbation_column_name} == '{target_perturbation_column_value}'"
            ),
        )

    def mahalanobis_distance(self, arguments: dict):
        """Workflow function: Add a column for the Mahalanobis distance to target perturbations.

        Designed to be called from a workflow, calculates the Mahalanobis distance
        in feature space.

        Parameters
        ----------
        arguments : dict, should contain:
            target_dataset
                Index or name of dataset which should be used in the measurment.
                If absent, then behaviour is as if -1 is supplied, indicating
                the last added dataset.
            target_perturbation_column_name
                normally a 'control' column
            target_perturbation_column_value:
                value to be found in the column defined previously.
            output_column_label:
                Output column for the measurement. If this is missing, then it is set to target_perturbation_column_value.

        """
        if "output_column_label" not in arguments:
            arguments["output_column_label"] = arguments["target_perturbation_column_value"]
        target_perturbation_column_name = arguments["target_perturbation_column_name"]
        target_perturbation_column_value = arguments["target_perturbation_column_value"]
        output_column_label = arguments["output_column_label"]
        ds = self.phe[arguments.get("target_dataset", -1)]
        ds.df[output_column_label] = phenonaut.metrics.mahalanobis(
            ds.data,
            ds.df.query(
                f"{target_perturbation_column_name} == '{target_perturbation_column_value}'"
            ),
        )

    def manhattan_distance(self, arguments: dict):
        """Workflow function: Add a column for the Manhattan distance to target perturbation.

        Designed to be called from a workflow, calculates the Manhattan distance
        in feature space. Also known as the cityblock distance.

        Parameters
        ----------
        arguments : dict, should contain:
            target_dataset
                Index or name of dataset which should be used in the measurment.
                If absent, then behaviour is as if -1 is supplied, indicating
                the last added dataset.
            target_perturbation_column_name
                normally a 'control' column
            target_perturbation_column_value:
                value to be found in the column defined previously.
            output_column_label:
                Output column for the measurement. If this is missing, then it is set to target_perturbation_column_value.

        """
        if "output_column_label" not in arguments:
            arguments["output_column_label"] = arguments["target_perturbation_column_value"]
        target_perturbation_column_name = arguments["target_perturbation_column_name"]
        target_perturbation_column_value = arguments["target_perturbation_column_value"]
        output_column_label = arguments["output_column_label"]
        ds = self.phe[arguments.get("target_dataset", -1)]
        ds.df[output_column_label] = phenonaut.metrics.manhattan(
            ds.data,
            ds.df.query(
                f"{target_perturbation_column_name} == '{target_perturbation_column_value}'"
            ),
        )

    def cityblock_distance(self, arguments: dict):
        """Workflow function: Add a column for the cityblock distance to a target perturbation.

        Designed to be called from a workflow, calculates the cityblock distance
        in feature space. Also known as the Manhattan distance.

        Parameters
        ----------
        arguments : dict, should contain:
            target_dataset
                Index or name of dataset which should be used in the measurment. If absent, then behaviour is as if -1 is supplied, indicating the last added dataset.
            target_perturbation_column_name
                normally a 'control' column
            target_perturbation_column_value:
                value to be found in the column defined previously.
            output_column_label:
                Output column for the measurement. If this is missing, then it is set to target_perturbation_column_value.

        """
        if "output_column_label" not in arguments:
            arguments["output_column_label"] = arguments["target_perturbation_column_value"]
        target_perturbation_column_name = arguments["target_perturbation_column_name"]
        target_perturbation_column_value = arguments["target_perturbation_column_value"]
        output_column_label = arguments["output_column_label"]
        ds = self.phe[arguments.get("target_dataset", -1)]
        ds.df[output_column_label] = phenonaut.metrics.manhattan(
            ds.data,
            ds.df.query(
                f"{target_perturbation_column_name} == '{target_perturbation_column_value}'"
            ),
        )

    def rename_column(self, arguments: dict):
        """Workflow function: Rename column

        Designed to be called from a workflow, ranames a single, or multiple
        columns. The arguments dictionary should contain key:value pairs, where
        the key is the old column name and the value is the new column name.

        Parameters
        ----------
        arguments : dict
            Dictionary containing name_from:name_to key value pairs, which will
            cause the column named 'name_from' to be renamed 'name_to'.
            Multiple columns can be renamed in a single call, using multiple
            dictionary entries.

        Raises
        ------
        ValueError
            'arguments' was not a dictionary of type: str:str
        """
        if isinstance(arguments, str):
            message = f"The argument to rename_column/rename_columns was a string, not a dictionary, if using a YAML workflow to define the job, make sure there is a space after the key - eg. 'oldkey: newkey', rather than 'oldkey:newkey'"
            raise ValueError(message)
        if len(arguments.keys()) > 0:
            self.phe.ds.rename_columns(arguments)

    def rename_columns(self, arguments: dict):
        """Workflow function: Rename columns

        Designed to be called from a workflow, ranames a single, or multiple
        columns. The arguments dictionary should contain key:value pairs, where
        the key is the old column name and the value is the new column name.

        Parameters
        ----------
        arguments : dict
            Dictionary containing name_from:name_to key value pairs, which will
            cause the column named 'name_from' to be renamed 'name_to'.
            Multiple columns can be renamed in a single call, using multiple
            dictionary entries.

        Raises
        ------
        ValueError
            'arguments' was not a dictionary of type: str:str
        """
        self.rename_column(arguments)

    def add_well_id(self, arguments: dict):
        """Workflow function: Add well IDs

        Designed to be called from a workflow. Often, we would like to use well
        and column numbers to resolve a more traditional alpha-numeric WellID
        notation, such as A1, A2, etc. This can be achieved through calling this
        workflow function.

        If a dataset contains numerical row and column names, then they may be
        translated into standard letter-number well IDs. The arguments
        dictionary may contain the following keys, with their values denoted
        as bellow:

        numerical_column_name : str, optional
            Name of column containing numeric column number, if not supplied,
            then behaves as if "COLUMN".
        numerical_row_name : str, optional
            Name of column containing numeric column number, if not supplied,
            then behaves as if "ROW".
        plate_type : int, optional
            Plate type - note, at present, only 384 well plate format is
            supported, if not supplied, then behaves as if 384.
        new_well_column_name : str, optional
            Name of new column containing letter-number well ID, if not
            supplied, then behaves as if default "Well".
        add_empty_wells : bool, optional
            Should all wells from a plate be inserted, even when missing from
            the data, if not supplied, then behaves as if False.
        plate_barcode_column : str, optional
            Multiple plates may be in a dataset, this column contains their
            unique ID, if not supplied, then bahaves as if None.
        no_sort : bool, optional
            Do not resort the dataset by well ID, if not supplied, then behaves
            as if False

        Parameters
        ----------
        arguments : dict
            Dictionary containing arguments to the Dataset.add_well_id function,
            see API documentation for further details, or function help.

        """
        self.phe.ds.add_well_id(**arguments)

    def copy_column(self, arguments: dict):
        """Workflow function: copy a dataset column

        Designed to be called from a workflow, copys the values of one column
        within a dataset to another. The arguments dictionary can contain 'to'
        and 'from' keys with values for column names, or alternatively, simply
        from:to key-value pairs denoting how to perform the copy operation.

        Parameters
        ----------
        arguments : dict
            Options for the command. Should include either:
                1.  dictionary with keys "to" and "from", with item names related to the
                columns that should be used.

                2.dictionary with the form {from_column:to_column}, which will copy the
                column with title from_column to to_column

            Note, if any dictionary items (to) are lists, then multiple copies will
            be made.

        Raises
        ------
        KeyError
            Column was not found in the Pandas DataFrame.
        """
        from_column = arguments.pop("from", None)
        to_column = arguments.pop("to", None)
        if any([from_column, to_column]):
            if not all([from_column, to_column]):
                message = "Only one of 'to' or 'from' found as keys, copy_columns needs a from and to argument, alternatively, give a dictionary key pair of colA:colB to copy colA to colB"
                raise KeyError(message)
            arguments[from_column] = to_column
        for from_column, to_column in arguments.items():
            if not isinstance(to_column, list):
                to_column = [to_column]
            for destination in to_column:
                print(f"Copying {from_column} to {destination}")
                self.phe.ds.df[destination] = self.phe.ds.df[from_column]

    def if_blank_also_blank(self, arguments: dict):
        """Workflow function: if column is empty, also blank

        Designed to be called from a workflow, often it is required to clean or
        remove rows not needed for inclusion into further established pipelines/
        workflows. This workflow function allows the ability to remove values
        from a column on the condition that onther column is empty.

        Parameters
        ----------
        arguments : dict
            Dictionary containing the following key:value pairs:

            query_column
                value is the name of the column to perform the query on.
            regex_query
                value is a boolean denoting if the query column value
                should be matched using a regular expression. If omitted, then
                behaves as if present and False.
            target_column
                value is a string, denoting the name of the column
                which should be blanked.
            target_columns
                value is a list of strings, denoting the names of columns which should be blanked.
            regex_targets
                value is a boolean denoting if the target column or multiple target columns defined in target_columns should be matched using a regular expression. If absent, then behaves as if False was supplied.

        Raises
        ------
        KeyError
            'query_column' not found in arguments dictionary
        IndexError
            Multiple columns matched query_column using the regex
        KeyError
            No target columns found for if_blank_also_blank, use target_column
            keys
        """

        query_column = arguments.pop("query_column", None)
        if query_column is None:
            message = f"if_blank_also_blank needs a 'query_column' argument"
            raise KeyError(message)
        if arguments.pop("regex_query", False):
            query_column = list(filter(re.compile(query_column).match, self.phe.df.columns.values))
            if len(query_column) == 1:
                query_column = query_column[0]
            else:
                message = "Multiple columns matched query_column using the regex"
                raise IndexError(message)

        target_columns = []
        for tc in ["target_column", "target_columns"]:
            if tc in arguments:
                tmp_target_cols = arguments.pop(tc)
                if isinstance(tmp_target_cols, str):
                    tmp_target_cols = [tmp_target_cols]
                target_columns.extend(tmp_target_cols)
        if len(target_columns) == 0:
            message = "No target columns found for if_blank_also_blank, use target_column keys"
            raise KeyError(message)

        if arguments.pop("regex_targets", False):
            expanded_target_columns = []
            for pattern in target_columns:
                for match in list(filter(re.compile(pattern).match, self.phe.df.columns.values)):
                    if match not in expanded_target_columns:
                        expanded_target_columns.append(match)
            target_columns = expanded_target_columns

        print(f"{target_columns=}  {query_column}")
        self.phe.df.loc[self.phe.df[query_column].isnull(), target_columns] = ""

    def pca(self, arguments: dict):
        """Workflow function: Perform PCA dimensionality reduction technique.

        Designed to be called from a workflow, performs the principal component
        dimensionality reduction technique. If no arguments are given or the
        arguments dictionary is empty, then 2D PCA is applied to the dataset
        with the highest index, equivalent of phe[-1], which is usually the last
        inserted dataset.

        Parameters
        ----------
        arguments : dict
            Dictionary of arguments to used to direct the PCA process, can
            contain the following keys and values

            target_dataset
                Index or name of dataset which should have the
                dimensionality reduction applied. If absent, then
                behaviour is as if -1 is supplied, indicating the last added
                dataset.
            ndims
                Number of dimensions to which the PCA should reduce the
                features. If absent, then defaults to 2.
            center_on_perturbation_id
                PCA should be recentered on the
                perturbation with ID. If absent, then defaults to None, and
                no centering is performed.
            center_by_median
                If true, then median of center_on_perturbation
                is used, if False, then the mean is used.
            fit_perturbation_ids
                PCA may be fit to only the included IDs,
                before the transform is applied to the whole dataset.
        """
        ndims = arguments.get("ndims", 2)
        center_on_perturbation = arguments.get("center_on_perturbation", None)
        from phenonaut.transforms import PCA

        transformer = PCA()
        transformer(
            self.phe[arguments.get("target_dataset", -1)],
            arguments.get("ndims", 2),
            center_on_perturbation_id=arguments.get("center_on_perturbation_id"),
            center_by_median=arguments.get("center_by_median", True),
            fit_perturbation_ids=arguments.get("fit_perturbation_ids"),
        )

    def umap(self, arguments: dict):
        """Workflow function: Perform UMAP dimensionality reduction technique.

        Designed to be called from a workflow, performs the UMAP dimensionality
        reduction technique. If no arguments are given or the arguments
        dictionary is empty, then 2D UMAP is applied to the dataset with the
        highest index, equivalent of phe[-1], which is usually the last inserted
        dataset.

        Parameters
        ----------
        arguments : dict
            Dictionary of arguments used to direct the UMAP transform function.
            Can contain the following keys and values.

            target_dataset
                Index or name of dataset which should have the
                dimensionality reduction applied. If absent, then
                behaviour is as if -1 is supplied, indicating the last added
                dataset.
            ndims
                Number of dimensions to which the UMAP should reduce the
                features. If absent, then defaults to 2.
            center_on_perturbation_id
                UMAP should be recentered on the
                perturbation with ID. If absent, then defaults to None, and
                no centering is performed.
            center_by_median
                If true, then median of center_on_perturbation
                is used, if False, then the mean is used.
        """
        ndims = arguments.get("ndims", 2)
        center_on_perturbation = arguments.get("center_on_perturbation", None)
        from phenonaut.transforms import UMAP

        transofmer = UMAP()
        transofmer(
            self.phe[arguments.get("target_dataset", -1)],
            arguments.get("ndims", 2),
            center_on_perturbation_id=arguments.get("center_on_perturbation_id"),
            center_by_median=arguments.get("center_by_median", True),
        )

    def tsne(self, arguments: dict):
        """Workflow function: Perform t-SNE dimensionality reduction technique.

        Designed to be called from a workflow, performs the t-SNE dimensionality
        reduction technique. If no arguments are given or the arguments
        dictionary is empty, then 2D t-SNE is applied to the dataset with the
        highest index, equivalent of phe[-1], which is usually the last inserted
        dataset.



        Parameters
        ----------
        arguments : dict
            Dictionary of arguments to used to direct the t-SNE process, can
            contain the following keys and values:

            target_dataset
                Index or name of dataset which should have the
                dimensionality reduction applied. If absent, then
                behaviour is as if -1 is supplied, indicating the last added
                dataset.
            ndims
                number of dimensions to which the t-SNE should reduce
                the features. If absent, then defaults to 2.
            center_on_perturbation_id
                tSNE should be recentered on the
                perturbation with ID.If absent, then defaults to None, and
                no centering is performed.
            center_by_median
                If true, then median of center_on_perturbation
                is used, if False, then the mean is used.
        """
        ndims = arguments.get("ndims", 2)
        center_on_perturbation = arguments.get("center_on_perturbation", None)
        from phenonaut.transforms import TSNE

        transofmer = TSNE()
        transofmer(
            self.phe[arguments.get("target_dataset", -1)],
            arguments.get("ndims", 2),
            center_on_perturbation_id=arguments.get("center_on_perturbation_id"),
            center_by_median=arguments.get("center_by_median", True),
        )

    def set_perturbation_column(self, arguments: dict):
        """Workflow function: Set the perturbation column

        Designed to be called from a workflow, the perturbation column can be
        set on a dataset to help with plotting/scatters.

        Parameters
        ----------
        arguments : dict
            target_dataset:
                Index or name of dataset within which we wish to set the perturbation
                column. If absent, then behaviour is as if -1 is supplied, indicating
                the last added dataset.
            column:
                str type giving the new column name which will be set to mark
                perturbations.

        """
        if "column" not in arguments:
            raise KeyError("Must supply a 'column' key in arguments dictionary")
        self.phe[arguments.get("target_dataset", -1)].perturbation_column = arguments["column"]

    def VIF_filter_features(self, arguments: dict):
        """Workflow function: Perform VIF feature filter

        Designed to be called from a workflow, performs variance inflation
        factor (VIF) filtering on a dataset, removing features which are not
        detrimental to capturing variance.
        More information available:
        https://en.wikipedia.org/wiki/Variance_inflation_factor

        This can be a computationally expensive process as the number of
        linear regressions required to be run is almost N^2 with features.

        Parameters
        ----------
        arguments : dict
            - target_dataset:
                Index or name of dataset which should have variance
                inflation filter applied. If absent, then behaviour is as if -1
                is supplied, indicating the last added dataset.

            - vif_cutoff:
                float or int indicating the VIF cutoff to apply. A good
                balance and value often used is 5.0.  If this key:value pair is
                absent, then behaviour is as if 5.0 was supplied.

            - min_features:
                removal of too many features can be detrimental.
                Setting this value sets a lower limit on the number of features
                which must remain.  If absent, then behaviour is as if a value
                of 2 was given.

            - drop_columns:
                value is a boolean, denoting if columns should be
                dropped from the data table, as well as being removed from
                features. If not supplied, then the behaviour is as if False was
                supplied.


        """
        from phenonaut.transforms.preparative import VIF

        vif = VIF()
        ds_id = arguments.pop("target_dataset", -1)
        vif(self.phe[ds_id], **arguments)

    def filter_correlated_features(self, arguments: dict):
        """Workflow function: Perform filter of highly correlated features

        Designed to be called from a workflow, performs filtering of highly
        correlated features (as calculated by Pearson correlation coefficient)
        by either by removal of features correlated above a given theshold, or
        uses the iterative removal of features with the highest R^2 against
        another feature.  The arguments dictionary should contain a threshold or
        n key:value pair, not both.  A key of threshold and float value defines
        the correlation above which, features should be removed. If the n key is
        present, then features are iteratively removeduntil n features remain.

        Parameters
        ----------
        arguments : dict

            target_dataset
                Index or name of dataset which should have features
                filtered. If absent, then behaviour is as if -1 is supplied,
                indicating the last added dataset.
            threshold
                If this key is present, then it activates threshold
                mode, where by calculated correlations, above which should be
                removed. A good value for this threshold is 0.9.
            n
                If this key is present, then the number of features to keep is
                defined this way. The process works through iteratively removing
                features ordered by the most correlated until the number of
                features is equal to n. If threshold is also present, then n
                acts as a minimum number of features and feature removal will
                stop, no matter the correlations present in the dataset.
            drop_columns : bool, optional
                If drop columns is True, then not only will features be removed
                from the dataset features list, but the columns for these
                features will be removed from the dataframe. If absent, then the
                behaviour is as if False was supplied as a value to this
                key:value pair.

        """
        from phenonaut.transforms.preparative import RemoveHighlyCorrelated

        rhc = RemoveHighlyCorrelated()
        ds_id = arguments.pop("target_dataset", -1)
        rhc(self.phe[ds_id], **arguments)

    def filter_correlated_and_VIF_features(self, arguments: dict):
        """Workflow function: Filter features by highly correlated then VIF.

        Designed to be called from a workflow, Ideally, VIF would be applied to
        very large datasets.  Due to the almost n^2 number of linear regression
        required as features increase, this is not possible on datasets with a
        large number of features - such as methylation datasets.  We therefore
        must use other methods to reduce the features to a comfortable level
        allowing VIF to be performed. This class calculates correlations between
        all features and iteratively (Pearson correlation coefficient), removing
        features with the highest R^2 against another feature. Once the number
        of featurs is reduced to a level suitable for VIF, VIF is performed.

        More information available:
        https://en.wikipedia.org/wiki/Variance_inflation_factor

        Parameters
        ----------
        arguments : dict
            - target_dataset:
                Index or name of dataset which should have features
                filtered. If absent, then behaviour is as if -1 is supplied,
                indicating the last added dataset.
            - n_before_vif :
                Number of features to remove before applying VIF.
                This is required when dealing with large datasets which would be
                too time consuming to process entirely with VIF.
                Features are removed iteratively, selecting the most correlated
                features and removing them. If this key:value pair is absent,
                then it is as if the value of 1000 has been supplied.
            - vif_cutoff :
                The VIF cutoff value, above which features are
                removed. Features with VIF scores above 5.0 are considered
                highly correlated. If not supplied, then behaviour is as if a
                value of 5.0 was supplied.
            - drop_columns :
                If drop columns is True, then not only will features
                be removed from the dataset features list, but the columns for
                these features will be removed from the dataframe. If absent,
                then behaviour is as if False was supplied.


        """
        from phenonaut.transforms.preparative import RemoveHighestCorrelatedThenVIF

        rhctv = RemoveHighestCorrelatedThenVIF()
        ds_id = arguments.pop("target_dataset", -1)
        rhctv(self.phe[ds_id], **arguments)

    def scatter(self, arguments: dict):
        """Workflow function: Make scatter plot.

        Designed to be called from a workflow, produce a scatter plot from a
        dataframe.

        Parameters
        ----------
        arguments : dict

            target_dataset
                Index or name of dataset which should have features
                plotted. If absent, then behaviour is as if -1 is supplied,
                indicating the last added dataset.
            figsize
                A tuple denoting the target output size in inches (!), if
                absent, then the default of (8,6) is used.
            title
                Title for the plot, if absent, then the default "2D scatter"
                is used.
            peturbations
                Can be a list of peturbations - as denoted in the
                perturbations column of the dataframe to include in the plot.
                If absent, then all perturbations are included.
            destination
                Output location for the PNG - required field.  An error
                will be thrown if omitted.

        """

        if "destination" not in arguments:
            raise KeyError(
                "The key 'destination' with a location to write the output PNG must be present within the arguments dictionary"
            )
        from phenonaut.output import Scatter

        ds_id = arguments.get("target_dataset", -1)
        figsize = arguments.get("figsize", (8, 6))
        title = arguments.get("title", "2D scatter")
        perturbations = arguments.get("perturbations", None)

        destination = arguments.get("destination")
        scatter = Scatter(figsize=figsize, title=title)
        scatter.add(self.phe[ds_id], perturbations=perturbations)
        scatter.save_figure(destination)


def predict(self, arguments: dict):
    """Workflow function: predict


    Profile predictors in their ability to predict a given target.

    Phenonaut provides functionality to profile the performance of multiple predictors against
    multiple views of data. This is exemplified in the TCGA example used in the Phenonaut paper -
    see Example 1 - TCGA for a full walkthrough of applying this functionality to The Cancer Genome Atlas.
    With a given ‘target’ for prediction which is in the dataset, predict selects all appropriate predictors
    (classifiers for classification, regressors for regression and multiregressors for multi regression/view targets).
    Then, enumerating all views of the data and all predictors, hyperparameter optimisation coupled with 5-fold cross
    validation using Optuna is employed, before finally testing the best hyperparameter sets with retained test sets.
    This process is automatic and requires the data, and a prediction target. Output from this process is extensive
    and it may take a long time to complete, depending on the characteristics of your input data. Writen output from
    the profiling process consists of performance heatmaps highlighting best view/predictor combinations in bold,
    boxplots for each view combination and a PPTX presentation file allowing easy sharing of data, along with machine
    readable CSV and JSON results.

    For each unique view combination and predictor, perform the following:

        - Merge views and remove samples which do not have features across currently needed views.

        - Shuffle the samples.

        - Withhold 20% of the data as a test set, to be tested against the trained and hyperparameter optimised predictor.

        - Split the data using 5-fold cross validation into train and validation sets.

        - For each fold, perform Optuna hyperparameter optimisation for the given predictor using the train sets, using hyperparameters described by the default predictors for classification, regression and multiregression.

    Parameters
    ----------

    arguments : dict

        output_directory
            Directory into which profiling output (boxplots, heatmaps, CSV, JSON and PPTX should be written).

        dataset_combinations
            If multiple datasets are already loaded, then lists of ‘views’ may be specified for exploration. If None, or this argument is absent, then all combinations of available views/Datasets are enumerated and used.

        target
            The prediction target, denoted by a column name given here which exists in loaded datasets.

        n_splits
            Number of splits to use in the N-fold cross validation, if absent, then the default of 5 is used.

        n_optuna_trials
            Number of Optuna trials for hyperparameter optimisation, by default 20. This drastically impacts runtime, so if things are taking too long, you may wish to lower this number. For a more thorough exploration of hyperparameter space, increase this number.

        optuna_merge_folds
            By default, each fold has hyperparameters optimised and the trained predictor with parameters reported. If this optuna_merge_folds is true, then each fold is trained on and and hyperparameters optimised across folds (not per-fold). Setting this to False may be useful depending on the intended use of the predictor. It is believed that when False, and parameters are not optimised across folds, then more accurate prediction variance/accuracy estimates are produced. If absent, behaves as if false.

        test_set_fraction
            When optimising a predictor, by default a fraction of the total data is held back for testing, separate from the train-validation splits. This test_set_fraction controls the size of this split. If absent, then the default value of 0.2 is used.

    """

    if "destination" not in arguments:
        raise KeyError(
            "The key 'destination' with a location to write the output PNG must be present within the arguments dictionary"
        )
    from phenonaut.predict import profile

    predict(**arguments)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

"""
Phenonaut workflows are an alternative way of running phenonaut transforms allowing
the processing of data as defined in YAML or JSON workflow files.
"""
from phenonaut.workflow import Workflow
from pathlib import Path
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a Phenonaut workflow")
    parser.add_argument("workflow_path", help="YAML or JSON file containing Phenonaut workflow")
    args = parser.parse_args()

    workflow_path = Path(args.workflow_path)

    ph0_workflow = Workflow(workflow_path)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor, GradientBoostingRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.svm import SVR
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neural_network import MLPRegressor
from phenonaut.predict.predictor_dataclasses import (
    PhenonautPredictor,
    HyperparameterInt,
    HyperparameterFloat,
    HyperparameterCategorical,
    HyperparameterLog,
)

default_regressors = [
    PhenonautPredictor(
        "Random Forest",
        RandomForestRegressor,
        [
            HyperparameterCategorical("max_depth", [5, 10, 20, None]),
            HyperparameterInt("n_estimators", 50, 200),
            HyperparameterCategorical("max_features", [None, "auto", "log2"]),
        ],
        constructor_kwargs={"n_jobs": -1},
    ),
    PhenonautPredictor(
        "Linear SVR",
        SVR,
        [
            HyperparameterCategorical("kernel", ["linear"]),
            HyperparameterLog("C", 0.001, 10),
            HyperparameterCategorical("gamma", ["scale", "auto"]),
        ],
        dataset_size_cutoff=1000,
    ),
    PhenonautPredictor(
        "RBF SVR",
        SVR,
        [
            HyperparameterCategorical("kernel", ["rbf"]),
            HyperparameterLog("C", 0.001, 10),
            HyperparameterCategorical("gamma", ["scale", "auto"]),
        ],
        dataset_size_cutoff=1000,
    ),
    PhenonautPredictor(
        "AdaBoost Regressor", AdaBoostRegressor, HyperparameterInt("n_estimators", 50, 200)
    ),
    PhenonautPredictor("Gradient Boosting Regressor", GradientBoostingRegressor),
    PhenonautPredictor("Decision Tree Regressor", DecisionTreeRegressor),
    PhenonautPredictor(
        "Gaussian Process Regressor",
        GaussianProcessRegressor,
        constructor_kwargs={"copy_X_train": False},
        embed_in_results=False,
    ),
    PhenonautPredictor(
        "KNeighbors Regressor",
        KNeighborsRegressor,
        HyperparameterInt("n_neighbors", 2, 10),
        constructor_kwargs={"n_jobs": -1},
    ),
    PhenonautPredictor(
        "SciKit NeuralNet Regressor",
        MLPRegressor,
        [
            HyperparameterLog("alpha", 0.00001, 0.01),
            HyperparameterCategorical("solver", ["adam", "sgd"]),
            HyperparameterCategorical("max_iter", [1000]),
            HyperparameterCategorical("learning_rate", ["constant", "adaptive"]),
        ],
    ),
]



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from collections import namedtuple
from pathlib import Path
import pathlib
from typing import List, NamedTuple, Optional, Union
from ..data import Dataset
import pandas as pd
from .base import PackagedDataset
import gzip, tarfile
import shutil
from sklearn.datasets import load_breast_cancer

import pandas as pd
import numpy as np
import h5py
import datetime


class BreastCancer(PackagedDataset):
    """Breast Cancer Dataset from scikit-learn

    This PackagedDataset provides the Breast Cancer dataset from
    scikit-learn. This is also known as the Breast cancer Wisconsin
    (diagnostic) dataset.
    See the scikit-learn user guide for more information:
    https://scikit-learn.org/stable/datasets/toy_dataset.html#breast-cancer-dataset
    Original dataset information available at:
    https://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Diagnostic)

    Contains 569 unique samples breast cancer fine needle aspirates, each
    with 30 features, and one target of 0 or 1, denoting benign or
    malignant respectively.

    Parameters
    ----------
    root : Union[Path, str]
        Local directory containing the prepared dataset. If the dataset is
        not found here and the argument download=True is not given, then an
        error is raised. If download=True and the processed dataset is
        absent, then it is downloaded the directory pointed at by the
        'raw_data' argument detailed below. If raw_data_dir is a
        non-absolute path, such as a single directory,
        then it is created as a subdirectory of this root directory.
    download : bool, optional
        If true and the processed dataset is not found in the root
        directory, then the dataset is downloaded and processed.
        By default False.
    raw_data_dir : Optional[Union[Path, str]], optional
        If downloading and preparing the dataset, then a directory for the
        raw data may be specified. If a non-absolute location is given,
        then it is created in a subdirectory of the root directory
        specified as the first argument. Absolute paths may be used to
        place raw datafiles and intermediates in another location, such as
        scratch disks etc, by default Path("raw_data").
    """

    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
    ):
        """Breast Cancer Dataset from scikit-learn

        This PackagedDataset provides the Breast Cancer dataset from
        scikit-learn. This is also known as the Breast cancer Wisconsin
        (diagnostic) dataset.
        See the scikit-learn user guide for more information:
        https://scikit-learn.org/stable/datasets/toy_dataset.html#breast-cancer-dataset
        Original dataset information available at:
        https://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Diagnostic)

        Contains 569 unique samples breast cancer fine needle aspirates, each
        with 30 features, and one target of 0 or 1, denoting benign or
        malignant respectively.

        Parameters
        ----------
        root : Union[Path, str]
            Local directory containing the prepared dataset. If the dataset is
            not found here and the argument download=True is not given, then an
            error is raised. If download=True and the processed dataset is
            absent, then it is downloaded the directory pointed at by the
            'raw_data' argument detailed below. If raw_data_dir is a
            non-absolute path, such as a single directory,
            then it is created as a subdirectory of this root directory.
        download : bool, optional
            If true and the processed dataset is not found in the root
            directory, then the dataset is downloaded and processed.
            By default False.
        raw_data_dir : Optional[Union[Path, str]], optional
            If downloading and preparing the dataset, then a directory for the
            raw data may be specified. If a non-absolute location is given,
            then it is created in a subdirectory of the root directory
            specified as the first argument. Absolute paths may be used to
            place raw datafiles and intermediates in another location, such as
            scratch disks etc, by default Path("raw_data").
        """

        super().__init__(root, raw_data_dir)
        self.name = "Breast cancer dataset"
        self.processed_h5_file = self.root / "breast_cancer_Wisconsin_phenonaut.h5"

        # If the dataset is missing, get it
        self._call_if_file_missing(self.processed_h5_file, self._make, None)

        self.store = pd.HDFStore(self.processed_h5_file)

        self.register_ds_key("Breast_cancer_dataset")

        self.store.close()

    def _make(self, remove_intermediates=False):
        """Make IRIS HDF5 file

        Parameters
        ----------
        remove_intermediates : bool, optional
            If true, then the downloaded archive is removed, by default False.
        """
        df = load_breast_cancer(as_frame=True).frame
        h5_store = pd.HDFStore(self.processed_h5_file, "w", complevel=9)

        h5_store["/Breast_cancer_dataset"] = df

        h5_store["/creation_date"] = pd.Series(str(datetime.datetime.now()))

        h5_store.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        features = [f for f in list(df.columns) if f not in ["target"]]
        ds = Dataset("Breast_Cancer", df, {"features": features})
        store.close()
        return ds



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Optional, Union
import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F
from pathlib import Path
import numpy as np


class _MRegressorNN(nn.Module):
    def __init__(self, network_shape):
        super().__init__()
        self.layers = nn.ModuleList()
        for nodes_i in range(len(network_shape) - 1):
            self.layers.append(nn.Linear(network_shape[nodes_i], network_shape[nodes_i + 1]))
            self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(network_shape[-2], network_shape[-1]))
        self.layers.append(nn.Sigmoid())

    def forward(self, x):
        for layer in self.layers[:-1]:
            x = layer(x)
        return x


class MultiRegressorNN:
    def __init__(
        self,
        batch_size=128,
        learning_rate=1e-3,
        epochs=100,
        num_hidden_layers=1,
        hidden_layer_sizes: Optional[list[int]] = None,
        use_optimizer: str = "ADAM",
        seed: Optional[int] = None,
    ):
        if seed is not None:
            torch.manual_seed(seed)
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.num_hidden_layers = num_hidden_layers
        self.use_optimizer = use_optimizer
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.hidden_layer_sizes = hidden_layer_sizes
        if self.hidden_layer_sizes is None:
            self.network_shape = np.linspace(100, 100, num=self.num_hidden_layers + 2, dtype=int)
        else:
            self.network_shape = np.array(
                [
                    100,
                ]
                + self.hidden_layer_sizes
                + [100]
            )
        self.model = _MRegressorNN(self.network_shape).to(self.device)

    def fit(self, X, y):
        if self.network_shape[0] != X.shape[1] or self.network_shape[-1] != y.shape[1]:
            del self.model
            if self.hidden_layer_sizes is None:
                self.network_shape = np.linspace(
                    X.shape[1], y.shape[1], num=self.num_hidden_layers + 2, dtype=int
                )
            else:
                self.network_shape = np.array(
                    [
                        X.shape[1],
                    ]
                    + self.hidden_layer_sizes
                    + [y.shape[1]]
                )
            self.model = _MRegressorNN(self.network_shape).to(self.device)
        y = torch.tensor(y.astype(np.float32))
        X = torch.tensor(X.astype(np.float32))
        train_tensor = torch.utils.data.TensorDataset(X, y)
        train_loader = torch.utils.data.DataLoader(
            dataset=train_tensor, batch_size=self.batch_size, shuffle=True
        )

        optimizer = None
        if self.use_optimizer == "ADAM":
            optimizer = optim.Adam(self.model.parameters(), lr=self.learning_rate)
        elif self.use_optimizer == "SGD":
            optimizer = optim.SGD(self.model.parameters(), lr=self.learning_rate, momentum=0.9)
        else:
            raise ValueError(
                f"Valid options for use_optimizer argument are 'ADAM' or 'SGD', use_optimizer was {self.use_optimizer}"
            )
        criterion = nn.MSELoss()
        self.model.train()
        for epoch in range(self.epochs):  # loop over the dataset multiple times
            for i, data in enumerate(train_loader, 0):
                # get the inputs; data is a list of [inputs, labels]
                X_train, y_train = data
                # zero the parameter gradients
                optimizer.zero_grad()
                # forward + backward + optimize
                outputs = self.model(X_train.to(self.device))
                loss = torch.sqrt(criterion(outputs, y_train.to(self.device)))
                loss.backward()
                optimizer.step()

    def predict(self, X):
        self.model.eval()
        test_loss = 0
        X = torch.tensor(X.astype(np.float32)).to(self.device)
        with torch.no_grad():
            data = X.to(self.device)
            return self.model(data).to("cpu").numpy()



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from copy import deepcopy
from optparse import Option
from typing import Callable, Optional, Union
import pandas as pd
import numpy as np
import phenonaut
import optuna
from phenonaut.data.dataset import Dataset
from phenonaut.predict.predictor_dataclasses import PhenonautPredictionMetric, PhenonautPredictor
from phenonaut.predict.predict_utils import (
    PredictionType,
    get_y_from_target,
    get_prediction_type_from_y,
    get_df_from_optuna_db,
    get_best_predictor_dataset_df,
    get_metric,
    get_X_y,
)
import phenonaut.data
from .optuna_functions import run_optuna_opt, run_optuna_opt_merge_folds
from pathlib import Path
from phenonaut.utils import check_path


def profile(
    phe: phenonaut.Phenonaut,
    output_directory: str,
    dataset_combinations: Optional[Union[None, list[int], list[str]]] = None,
    target: Optional[Union[str, pd.Series, np.ndarray, phenonaut.data.Dataset]] = None,
    predictors: Optional[list[PhenonautPredictor]] = None,
    prediction_type: Optional[Union[str, PredictionType]] = None,
    n_splits: int = 5,
    seed: Optional[int] = None,
    optuna_db_path: Optional[Union[Path, str]] = None,
    optuna_db_protocol: str = "sqlite:///",
    n_optuna_trials=20,
    metric: Optional[Union[PhenonautPredictionMetric, str]] = None,
    no_output: bool = False,
    write_pptx: bool = True,
    optuna_merge_folds: bool = False,
    test_set_fraction: float = 0.2,
    save_models: bool = True,
):
    """Profile predictors in their ability to predict a given target.

    This predict.profile function operates on a Phenonaut object (optionally
    containing multiple Datasets) and a given or indicated prediction target.
    The data within the prediction target is examined and the prediction type
    determined from classification, regression, and multiregression/view
    prediction (prediction of 1 omics view from another). In the case of
    Example 1 in the Phenonaut paper, on TCGA, with the prediction target of
    “survives_1_year”, the data types within the metadata are examined and only
    two values found 0 (no) or 1 (yes). Classification is chosen. With no
    classifiers explicitly given as arguments to the profile function,
    Phenonaut selects all default classifiers. User supplied classifiers and
    predictors may be passed, including PyTorch neural networks and similar
    objects through wrapping in PhenonautPredictor dataclasses of any class that
    implements the fit and predict methods. See PhenonautPredictor API
    documentation for further information on including user defined and packaged
    predictors. With no views explicitly listed, all possible view combinations
    are selected. For TCGA, the four omics views allow 15 view combinations
    (4x singular, 6x double, 4x triple and 1x quad). For each unique view
    combination and predictor, perform the following:


    * Merge views and remove samples which do not have features across currently needed views.

    * Shuffle the samples.

    * Withhold 20% of the data as a test set, to be tested against the trained and hyperparameter optimised predictor.

    * Split the data using 5-fold cross validation into train and validation sets.

    * For each fold, perform Optuna hyperparameter optimisation for the given predictor using the train sets, using hyperparameters described by the default predictors for classification, regression and multiregression.


    Parameters
    ----------
    phe : phenonaut.Phenonaut
        A Phenonaut object containing Datasets for prediction on
    output_directory : str
        Directory into which profiling output (boxplots, heatmaps, CSV, JSON
        and PPTX should be written).
    dataset_combinations : Optional[Union[None, list[int], list[str]]], optional
        If the Phenonaut object contains multiple datasets, then tuples of
        'views' may be specified for exploration. If None, then all combinations
        of available views/Datasets are enumerated and used. By default None.
    target : Optional[Union[str, pd.Series, np.ndarray, phenonaut.data.Dataset]], optional
        The prediction target. May be an array-like structure for prediction
        from aligned views, a string denoting the column in which to find the
        prediction target data within one of the Phenonaut Datasets, or a
        pd.Series, by default None.
    predictors : Optional[list[PhenonautPredictor]], optional
        A list of PhenonautPredictors may be supplied to the function. If None,
        the all suitable predictors for the type of prediction problem that it
        is are selected (through loading of
        default_classifiers, default_regressors, or default_multiregressors from
        phenonaut/predict/default_predictors/ .
        By default None.,
    prediction_type : Optional[Union[str, PredictionType]], optional
        The type of prediction task like "classification", "regression" and
        "view", or the proper PredictionType enum. If None, then the prediction
        task is assigned through inspection of the target data types and values
        present.  By default, None.
    n_splits : int, optional
        Number of splits to use in the N-fold cross validation, by default 5.
    seed : Optional[int], optional
        Optionally set the seed used for random number generation.
        By default None.
    optuna_db_path : Optional[Union[Path, str]], optional
        Path to Optuna sqlite3 database file. If None, then a default filename
        will be assigned by Phenonaut. By default None.
    optuna_db_protocol : _type_, optional
        Protocol that Optuna should use for accessing its required persistent
        storage. By default "sqlite:///"
    n_optuna_trials : int, optional
        Number of Optuna trials for hyperparameter optimisation, by default 20
    metric : Optional[Union[PhenonautPredictionMetric, str]], optional
        Metric used for scoring, currently understands the shortcut strings:
        accuracy, accuracy_score
        mse, MSE, mean_squared_error
        rmse, RMSE, root_mean_squared_error
        AUROC, auroc, area_under_roc_curveby default None
    no_output : bool, optional
        If True, then no output is writen to disk. Hyperparameter optimisation
        is performed and the (usually) sqlite3 file written, without writing
        boxplot and heatmap images, CSVs, JSONS, and PPTX files
        By default False.
    write_pptx : bool, optional
        If True, then the output boxplots and heatmaps are written to a PPTX
        file for presentation/sharing of data. By default True.
    optuna_merge_folds : bool, optional
        By default, each fold has hyperparameters optimised and the trained
        predictor with parameters reported.  If this optuna_merge_folds is true,
        then each fold is trained on and and hyperparameters optimised across
        folds (not per-fold). Setting this to False may be useful depending on
        the intended use of the predictor. It is believed that when False, and
        parameters are not optimised across folds, then more accurate prediction
        variance/accuracy estimates are produced. By default False.
    test_set_fraction : float, optional
        When optimising a predictor, by default a fraction of the total data is
        held back for testing, separate from the train-validation splits. This
        test_set_fraction controls the size of this split. By default 0.2.
    save_models : bool, optional
        Save the trained models to pickle files for later use. By default False.

    """
    output_directory = check_path(output_directory, is_dir=True)

    # Set up optuna_db_path. Dont do any file checks on it - as a sql
    # database may be specified in combination with the optuna_db_protocol
    # argument.
    if optuna_db_path is None:
        optuna_db_path = output_directory / "optuna_scan.db"
    if predictors == [] and optuna_db_protocol == "sqlite:///" and not optuna_db_path.exists():
        raise ValueError(
            "predictors argument was empty, no optuna_db_path given and no 'optuna_scan.db' was found in the targeted output directory"
        )

    if dataset_combinations is None:
        dataset_combinations = phe.get_dataset_combinations()

    # Turn dataset_combination int tuples into named sets (may happen if user supplied)
    #    ((0,), (0,1))
    dataset_combinations = [
        [ds if isinstance(ds, str) else phe[ds].name for ds in comb]
        for comb in dataset_combinations
    ]

    # Set prediction_type - may be supplied as a string, matching one of the defined enum types.
    # If string, then assign the proper enum type. If None, then detect it.
    if isinstance(prediction_type, str):
        if prediction_type not in [pt.name for pt in PredictionType]:
            raise ValueError(
                f"prediction_type supplied as a string was not a valid option. It was '{prediction_type}', and should be one of:{'.'.join([pt.name for pt in PredictionType])}"
            )
        prediction_type = PredictionType[prediction_type]

    if prediction_type is None:
        # We need to get the prediction target - even before iterating over
        # datasets so that the prediction type can be set below.
        y = get_y_from_target(phe.datasets, target=target)
        prediction_type = get_prediction_type_from_y(y)
        print("Predicted prediction type = ", prediction_type.name)
        if metric is None:
            if prediction_type == PredictionType.classification:
                print("Setting metric to accuracy")
                metric = "accuracy"
            else:
                print("Setting metric to MSE")
                metric = "mse"
    metric = get_metric(metric)

    if n_optuna_trials > 0:
        # If not supplied, set the default predictors
        if predictors is None:
            if prediction_type == PredictionType.classification:
                from phenonaut.predict.default_predictors.classifiers import default_classifiers

                predictors = default_classifiers
            if prediction_type == PredictionType.regression:
                from .default_predictors.regressors import default_regressors

                predictors = default_regressors
            if prediction_type == PredictionType.view:
                from .default_predictors.multiregressors import default_multiregressors

                predictors = default_multiregressors

        for dataset_combination in dataset_combinations:
            print(f"Working on datasets: {dataset_combination=}")
            for predictor in predictors:
                X, y = get_X_y(phe, dataset_combination, target, predictor, prediction_type)

                if X is None or y is None:
                    continue
                if (
                    prediction_type == PredictionType.classification
                    and predictor.max_classes is not None
                    and predictor.max_classes < len(np.unique(y))
                ):
                    continue
                if (
                    predictor.dataset_size_cutoff is not None
                    and predictor.dataset_size_cutoff < len(y)
                ):
                    continue

                # Split out a test set, leaving train and validation in X and y
                shuffled_indexes = np.arange(y.shape[0], dtype=int)
                np.random.seed(seed)
                np.random.shuffle(shuffled_indexes)

                if isinstance(X, list):
                    X_test = [
                        Xv[shuffled_indexes[: int(y.shape[0] * test_set_fraction)]] for Xv in X
                    ]
                    X = [Xv[shuffled_indexes[int(y.shape[0] * test_set_fraction) :]] for Xv in X]
                else:
                    X_test = X[shuffled_indexes[: int(y.shape[0] * test_set_fraction)]]
                    X = X[shuffled_indexes[int(y.shape[0] * test_set_fraction) :]]
                y_test = y[shuffled_indexes[: int(y.shape[0] * test_set_fraction)]]
                y = y[shuffled_indexes[int(y.shape[0] * test_set_fraction) :]]

                if isinstance(target, Dataset):
                    target_dataset_name = target.name
                else:
                    target_dataset_name = None

                if optuna_merge_folds:
                    run_optuna_opt_merge_folds(
                        X,
                        X_test,
                        y,
                        y_test,
                        prediction_type=prediction_type,
                        predictor=predictor,
                        metric=metric,
                        n_optuna_trials=n_optuna_trials,
                        phe_name=phe.name,
                        dataset_combination=dataset_combination,
                        optuna_db_path=optuna_db_path,
                        n_splits=n_splits,
                        seed=seed,
                        optuna_db_protocol=optuna_db_protocol,
                        target_dataset_name=target_dataset_name,
                    )
                else:
                    run_optuna_opt(
                        X,
                        X_test,
                        y,
                        y_test,
                        prediction_type=prediction_type,
                        predictor=predictor,
                        metric=metric,
                        n_optuna_trials=n_optuna_trials,
                        phe_name=phe.name,
                        dataset_combination=dataset_combination,
                        optuna_db_path=optuna_db_path,
                        n_splits=n_splits,
                        seed=seed,
                        optuna_db_protocol=optuna_db_protocol,
                        target_dataset_name=target_dataset_name,
                    )

    if no_output:
        return

    # Write all scan results to CSV and JSON
    df_all = get_df_from_optuna_db(
        output_directory / "optuna_scan.db",
        csv_output_filename=output_directory / "optuna_scan_all.csv",
        json_output_filename=output_directory / "optuna_scan_all.json",
    )

    # Write best results to CSV and JSON
    df_best = get_df_from_optuna_db(
        output_directory / "optuna_scan.db",
        csv_output_filename=output_directory / "optuna_scan_best.csv",
        json_output_filename=output_directory / "optuna_scan_best.json",
        get_only_best_per_study=True,
    )

    # At this stage, we have written out CSV/JSON files.  Visualisations are sensitive to extreme values,
    # in the case of the GaussianProcess Regressor, it performs teribly in predicting Iris stem length, with an
    # RMSE of ~700 compared to ~0.1 for the rest of the predictors. We may therefore consider removing predictors
    # from visualisations which have score a certain number of StdDev from mean.
    # Experimentation focused on removing items 5x standard deviations away from the mean, however, this is
    # currently left out and remains unimplemented and untested.
    # df_all=df_all[np.abs(np.array((df_all.score-df_all.score.mean())) <= 5*(df_all.score.quantile(0.25)-df_all.score.quantile(0.75)))]
    # df_best=df_best[np.abs(np.array((df_best.score-df_best.score.mean())) <= 5*(df_best.score.quantile(0.25)-df_best.score.quantile(0.75)))]
    if optuna_merge_folds:
        best_with_folds_as_scores = []
        df_best = df_best.reset_index()  # make sure indexes pair with number of rows

        # When plotting the best predictors, we want boxplots to contain the fold results.
        # Therefore we create a new dataframe containing KFold instances of FoldScore,
        # which will be used to plot the results.
        for index, row in df_best.iterrows():
            for fold in row[df_best.columns.str.startswith("FoldScore_")].values:
                tmp_dict = row.to_dict()
                tmp_dict["FoldScore"] = fold
                best_with_folds_as_scores.append(tmp_dict)
        df_best = pd.DataFrame(best_with_folds_as_scores)
        best_predictor_dataset_heatmap = get_best_predictor_dataset_df(df_best)
        best_predictor_std_dataset_heatmap = get_best_predictor_dataset_df(
            df_best, column_containing_values="Std_KFold"
        )
        from phenonaut.output.heatmap import write_heatmap_from_df

        write_heatmap_from_df(
            best_predictor_dataset_heatmap.dropna(how="all"),
            "",
            output_file=output_directory / "best_predictor_ds_heatmap.png",
            standard_deviations_df=best_predictor_std_dataset_heatmap.dropna(how="all"),
            axis_labels=("Predictors", "Views"),
            transpose=False,
            lower_is_better=metric.lower_is_better,
        )

        from phenonaut.output.boxplot import write_boxplot_to_file

        boxplot_best_paths = []
        for dataset in df_best.dataset.unique():
            boxplot_best_paths.append(
                output_directory / f"boxplot_{dataset.replace(':', '_')}_best.png"
            )
            write_boxplot_to_file(
                df_best.query(f"dataset == '{dataset}'"),
                "predictor_name",
                "FoldScore",
                boxplot_best_paths[-1],
                f"{dataset}, best trials, all folds",
                "Predictor",
                df_best.query(f"dataset == '{dataset}'")["metric_name"].unique()[0],
            )

    else:
        best_with_folds_as_scores = []
        df_best = df_best.reset_index()  # make sure indexes pair with number of rows
        mean_df = (
            df_best.groupby(["dataset", "predictor_name"])
            .mean()
            .reset_index()[["dataset", "predictor_name", "test_score"]]
            .dropna(axis=0, how="all")
            .dropna(axis=1, how="all")
        )
        sd_df = (
            df_best.groupby(["dataset", "predictor_name"])
            .std()
            .reset_index()[["dataset", "predictor_name", "test_score"]]
            .dropna(axis=0, how="all")
            .dropna(axis=1, how="all")
        )

        # When plotting the best predictors, we want boxplots to contain the fold results.
        # Therefore we create a new dataframe containing KFold instances of FoldScore,
        # which will be used to plot the results.
        # for index, row in mean_df.iterrows():
        #     tmp_dict = row.to_dict()
        #     tmp_dict["FoldScore"] = fold
        #     best_with_folds_as_scores.append(tmp_dict)
        # df_best = pd.DataFrame(best_with_folds_as_scores)
        # df_best.to_csv(output_directory/"b.csv")
        best_predictor_dataset_heatmap = get_best_predictor_dataset_df(
            mean_df, column_containing_values="test_score"
        )
        best_predictor_std_dataset_heatmap = get_best_predictor_dataset_df(
            sd_df, column_containing_values="test_score"
        )

        best_predictor_dataset_heatmap = best_predictor_dataset_heatmap.reindex(
            df_best.predictor_name.unique(), axis=1
        ).reindex(df_best.dataset.unique(), axis=0)
        best_predictor_std_dataset_heatmap = best_predictor_std_dataset_heatmap.reindex(
            df_best.predictor_name.unique(), axis=1
        ).reindex(df_best.dataset.unique(), axis=0)

        from phenonaut.output.heatmap import write_heatmap_from_df

        write_heatmap_from_df(
            best_predictor_dataset_heatmap.dropna(how="all"),
            "",
            output_file=output_directory / "best_predictor_ds_heatmap.png",
            standard_deviations_df=best_predictor_std_dataset_heatmap.dropna(how="all"),
            axis_labels=("Predictors", "Views"),
            transpose=False,
            lower_is_better=metric.lower_is_better,
        )

        from phenonaut.output.boxplot import write_boxplot_to_file

        boxplot_best_paths = []
        for dataset in df_best.dataset.unique():
            boxplot_best_paths.append(
                output_directory / f"boxplot_{dataset.replace(':', '_')}_best.png"
            )
            write_boxplot_to_file(
                df_best.query(f"dataset == '{dataset}'"),
                "predictor_name",
                "test_score",
                boxplot_best_paths[-1],
                f"{dataset}, best trials, all folds",
                "Predictor",
                df_best.query(f"dataset == '{dataset}'")["metric_name"].unique()[0],
            )
    if write_pptx:
        from phenonaut.output.pptx import PhenonautPPTX

        ppt = PhenonautPPTX(cover_subtitle=phe.name)
        # Make dataset name from the path (2 options here, ends with _best or _all)
        dfromp = (
            lambda f: str(f.stem).replace("_best", "").replace("boxplot_", "")
            if str(f.stem).endswith("_best")
            else str(f.stem).replace("_all", "").replace("boxplot_", "")
        )

        # Sort filenames first by alphabetical, then by length then iterate
        # Unneded: for f in sorted(boxplot_image_filenames, key=lambda x: (len(x.name), x.name)):
        ppt.add_image_slide(
            "Heatmap", output_directory / "best_predictor_ds_heatmap.png", width=15, height=15
        )
        for f in boxplot_best_paths:
            ppt.add_image_slide(f"{dfromp(f)}", f)
        ppt.save(output_directory / f"prediction_performance.pptx")
